SystemOrganization addCategory: #SoundMorphicUserInterface!

ImageMorph subclass: #SoundMorph
	instanceVariableNames: 'sound'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SoundMorph commentStamp: '<historical>' prior: 0!
Note: as of December 2000, this does not work. SoundMorph>>buildImage requires the sound to implement #volumeEnvelopeScaledTo: and as yet, no one does.!

!SoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:21'!
buildImage
	| scale env h imageColor |
	owner ifNil: [scale := 128@128]  "Default is 128 pix/second, 128 pix fullscale"
		ifNotNil: [scale := owner soundScale].
	env := sound volumeEnvelopeScaledTo: scale.
	self image: (ColorForm extent: env size @ env max).
	1 to: image width do:
		[:x | h := env at: x.
		image fillBlack: ((x-1)@(image height-h//2) extent: 1@h)].
	imageColor := #(black red orange green blue) atPin:
						(sound pitch / 110.0) rounded highBit.
	image colors: (Array with: Color transparent with: (Color perform: imageColor)).
! !

!SoundMorph methodsFor: 'initialization' stamp: 'jm 12/17/97 22:43'!
initialize

	super initialize.
	self sound: (FMSound pitch: 880.0 dur: 0.2 loudness: 0.8).
! !

!SoundMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 20:05'!
justDroppedInto: aMorph event: anEvent
	| relPosition |
	relPosition := self position - aMorph innerBounds topLeft.
	relPosition := (relPosition x roundTo: 8) @ relPosition y.
	self position: aMorph innerBounds topLeft + relPosition.
	sound copy play.
	^super justDroppedInto: aMorph event: anEvent! !

!SoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:08'!
reset
	sound reset! !

!SoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:08'!
sound
	^ sound! !

!SoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:08'!
sound: aSound
	sound := aSound copy.
	sound reset.
	self buildImage! !

PolygonMorph subclass: #EnvelopeLineMorph
	instanceVariableNames: 'editor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!EnvelopeLineMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:27'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color transparent! !

!EnvelopeLineMorph methodsFor: 'editing' stamp: 'ar 3/17/2001 14:38'!
dragVertex: ix event: evt fromHandle: handle
	| p |
	super dragVertex: ix event: evt fromHandle: handle.
	p := owner acceptGraphPoint: evt cursorPoint at: ix.
	self verticesAt: ix put: p.
! !

!EnvelopeLineMorph methodsFor: 'editing' stamp: 'ar 3/17/2001 14:31'!
dropVertex: ix event: evt fromHandle: handle
	| oldVerts |
	oldVerts := vertices.
	super dropVertex: ix event: evt fromHandle: handle.
	vertices = oldVerts ifFalse: [owner deletePoint: ix "deleted a vertex"]! !

!EnvelopeLineMorph methodsFor: 'editing' stamp: 'ar 3/17/2001 14:39'!
newVertex: ix event: evt fromHandle: handle
	"Install a new vertex if there is room."
	(owner insertPointAfter: ix) ifFalse: [^ self "not enough room"].
	super newVertex: ix event: evt fromHandle: handle.
	self verticesAt: ix+1 put: (owner acceptGraphPoint: evt cursorPoint at: ix+1).
! !

!EnvelopeLineMorph methodsFor: 'as yet unclassified' stamp: 'dgd 2/14/2003 20:17'!
vertices: verts borderWidth: bw borderColor: bc 
	super initialize.
	vertices := verts.
	
	borderWidth := bw.
	borderColor := bc.
	closed := false.
	arrows := #none.
	self computeBounds! !

AlignmentMorph subclass: #MIDIControllerMorph
	instanceVariableNames: 'channel controller midiPort lastValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!MIDIControllerMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:50'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'set channel' translated action: #setChannel:.
	aCustomMenu add: 'set controller' translated action: #setController:.
! !

!MIDIControllerMorph methodsFor: 'accessing' stamp: 'jm 9/28/1998 22:35'!
channel

	^ channel
! !

!MIDIControllerMorph methodsFor: 'accessing' stamp: 'jm 9/28/1998 22:40'!
channel: anInteger

	channel := anInteger.
	lastValue := nil.
	self updateLabel.
! !

!MIDIControllerMorph methodsFor: 'accessing' stamp: 'jm 9/28/1998 22:35'!
controller

	^ controller
! !

!MIDIControllerMorph methodsFor: 'accessing' stamp: 'jm 9/28/1998 22:40'!
controller: anInteger

	controller := anInteger.
	lastValue := nil.
	self updateLabel.
! !

!MIDIControllerMorph methodsFor: 'menu' stamp: 'jm 9/29/1998 09:06'!
controllerList
	"Answer a list of controller name, number pairs to be used in the menu."

	^ #((1 modulation)
		(2 'breath control')
		(7 volume)
		(10 pan)
		(11 expression)
		(92 'tremolo depth')
		(93 'chorus depth')
		(94 'celeste depth')
		(95 'phaser depth'))
! !

!MIDIControllerMorph methodsFor: 'menu' stamp: 'jm 9/29/1998 09:09'!
controllerName: controllerNumber
	"Answer a name for the given controller. If no name is available, use the form 'CC5' (CC is short for 'continuous controller')."

	self controllerList do: [:pair |
		pair first = controllerNumber ifTrue: [^ pair last]].
	^ 'CC', controllerNumber asString
! !

!MIDIControllerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:37'!
defaultBorderWidth
"answer the default border width for the receiver"
	^ 1! !

!MIDIControllerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:37'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color
		r: 0.484
		g: 0.613
		b: 0.0! !

!MIDIControllerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:37'!
initialize
"initialize the state of the receiver"
	| slider |
	super initialize.
""
	self listDirection: #topToBottom.
	self wrapCentering: #center;
		 cellPositioning: #topCenter.
	self hResizing: #shrinkWrap.
	self vResizing: #shrinkWrap.
	channel := 0.
	controller := 7.
	"channel volume"
	slider := SimpleSliderMorph new target: self;
				 actionSelector: #newSliderValue:;
				 minVal: 0;
				 maxVal: 127;
				 extent: 128 @ 10.
	self addMorphBack: slider.
	self
		addMorphBack: (StringMorph contents: 'Midi Controller').
	self updateLabel! !

!MIDIControllerMorph methodsFor: 'accessing' stamp: 'jm 9/28/1998 22:54'!
midiPort

	^ midiPort
! !

!MIDIControllerMorph methodsFor: 'accessing' stamp: 'jm 9/28/1998 22:55'!
midiPort: anInteger

	midiPort := anInteger.
! !

!MIDIControllerMorph methodsFor: 'other' stamp: 'jm 10/12/1998 16:02'!
newSliderValue: newValue
	"Send a control command out the MIDI port."

	| val |
	midiPort ifNil: [^ self].
	val := newValue asInteger.
	lastValue = val ifTrue: [^ self].
	lastValue := val.
	midiPort midiCmd: 16rB0 channel: channel byte: controller byte: val.
! !

!MIDIControllerMorph methodsFor: 'menu' stamp: 'RAA 6/12/2000 09:02'!
setChannel: evt
	| menu |
	menu := MenuMorph new.
	1 to: 16 do: [:chan |
		menu add: chan printString
			target: self
			selector: #channel:
			argumentList: (Array with: chan - 1)].

	menu popUpEvent: evt in: self world! !

!MIDIControllerMorph methodsFor: 'menu' stamp: 'RAA 6/12/2000 09:02'!
setController: evt
	| menu |
	menu := MenuMorph new.
	self controllerList do: [:pair |
		menu add: (pair last)
			target: self
			selector: #controller:
			argumentList: (Array with: pair first)].

	menu popUpEvent: evt in: self world! !

!MIDIControllerMorph methodsFor: 'other' stamp: 'jm 9/29/1998 09:10'!
updateLabel

	| label |
	(label := self findA: StringMorph) ifNil: [^ self].
	label contents: (self controllerName: controller), ', ch: ', (channel + 1) printString.
! !

AlignmentMorph subclass: #MidiInputMorph
	instanceVariableNames: 'midiPortNumber midiSynth instrumentSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!MidiInputMorph commentStamp: '<historical>' prior: 0!
I am the user interface for a simple software MIDI synthesizer that is driven by external MIDI input. I come with controls for a single MIDI channel (channel 1), but allow channel controls for additional MIDI channels to be added by the user. The volume, pan, and instrument of each channel can be controlled independently.
!

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'alain.plantec 2/8/2009 22:22'!
addChannel
	"Add a set of controls for another channel. Prompt the user for the channel number."

	| existingChannels newChannel candidates |
	existingChannels := Set new.
	1 to: 16 do: [:ch | (instrumentSelector at: ch) ifNotNil: [existingChannels add: ch]].
	candidates := (1 to: 16) select: [:ch | (existingChannels includes: ch) not].
	newChannel := UIManager default chooseFrom: (candidates collect: [:c | c printString]) values: candidates.
	newChannel ifNotNil: [self addChannelControlsFor: newChannel].
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
addChannelControlsFor: channelIndex

	| r divider col |
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: (self channelNumAndMuteButtonFor: channelIndex).
	r addMorphBack: (Morph new extent: 10@5; color: color).  "spacer"
	r addMorphBack: (self panAndVolControlsFor: channelIndex).

	divider := AlignmentMorph new
		extent: 10@1;
		borderWidth: 1;
		layoutInset: 0;
		borderColor: #raised;
		color: color;
		hResizing: #spaceFill;
		vResizing: #rigid.

	col := self lastSubmorph.
	col addMorphBack: divider.
	col addMorphBack: r.
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'gm 2/28/2003 00:00'!
atChannel: channelIndex from: aPopUpChoice selectInstrument: selection 
	| oldSnd name snd instSelector |
	oldSnd := midiSynth instrumentForChannel: channelIndex.
	(selection beginsWith: 'edit ') 
		ifTrue: 
			[name := selection copyFrom: 6 to: selection size.
			aPopUpChoice contentsClipped: name.
			(oldSnd isKindOf: FMSound) | (oldSnd isKindOf: LoopedSampledSound) 
				ifTrue: [EnvelopeEditorMorph openOn: oldSnd title: name].
			(oldSnd isKindOf: SampledInstrument) 
				ifTrue: [EnvelopeEditorMorph openOn: oldSnd allNotes first title: name].
			^self].
	snd := nil.
	1 to: instrumentSelector size
		do: 
			[:i | 
			(channelIndex ~= i and: 
					[(instSelector := instrumentSelector at: i) notNil 
						and: [selection = instSelector contents]]) 
				ifTrue: [snd := midiSynth instrumentForChannel: i]].	"use existing instrument prototype"
	snd ifNil: 
			[snd := (selection = 'clink' 
						ifTrue: 
							[(SampledSound samples: SampledSound coffeeCupClink samplingRate: 11025)]
						ifFalse: [(AbstractSound soundNamed: selection) ])copy ].
	midiSynth instrumentForChannel: channelIndex put: snd.
	(instrumentSelector at: channelIndex) contentsClipped: selection! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
channelNumAndMuteButtonFor: channelIndex

	| muteButton instSelector r |
	muteButton := SimpleSwitchMorph new
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		offColor: color;
		color: color;
		label: 'Mute';
		target: midiSynth;
		actionSelector: #mutedForChannel:put:;
		arguments: (Array with: channelIndex).
	instSelector := PopUpChoiceMorph new
		extent: 95@14;
		contentsClipped: 'oboe1';
		target: self;
		actionSelector: #atChannel:from:selectInstrument:;
		getItemsSelector: #instrumentChoicesForChannel:;
		getItemsArgs: (Array with: channelIndex).
	instSelector arguments:
		(Array with: channelIndex with: instSelector).
	instrumentSelector at: channelIndex put: instSelector.

	r := self makeRow
		hResizing: #rigid;
		vResizing: #spaceFill;
		extent: 70@10.
	r addMorphBack:
		(StringMorph
			contents: channelIndex printString
			font: (TextStyle default fontOfSize: 24)).
	channelIndex < 10
		ifTrue: [r addMorphBack: (Morph new color: color; extent: 19@8)]  "spacer"
		ifFalse: [r addMorphBack: (Morph new color: color; extent: 8@8)].  "spacer"
	r addMorphBack: instSelector.
	r addMorphBack: (AlignmentMorph newRow color: color).  "spacer"
	r addMorphBack: muteButton.
	^ r
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'jm 1/13/1999 07:33'!
closeMIDIPort

	midiSynth isOn ifTrue: [midiSynth stopMIDITracking].
	midiSynth closeMIDIPort.
! !

!MidiInputMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:38'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!MidiInputMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:28'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color veryLightGray! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'jm 1/6/1999 07:55'!
disableReverb: aBoolean

	aBoolean
		ifTrue: [SoundPlayer stopReverb]
		ifFalse: [SoundPlayer startReverb].
! !

!MidiInputMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self listDirection: #topToBottom;
	  wrapCentering: #center;
		 cellPositioning: #topCenter;
	  hResizing: #spaceFill;
	  vResizing: #spaceFill;
	  layoutInset: 3.
	midiPortNumber := nil.
	midiSynth := MIDISynth new.
	instrumentSelector := Array new: 16.
	self removeAllMorphs.
	self addMorphBack: self makeControls.
	self addMorphBack: (AlignmentMorph newColumn color: color;
			 layoutInset: 0).
	self addChannelControlsFor: 1.
	self extent: 20 @ 20! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
instrumentChoicesForChannel: channelIndex

	| names inst |
	names := AbstractSound soundNames asOrderedCollection.
	names := names collect: [:n |
		inst := AbstractSound soundNamed: n.
		(inst isKindOf: UnloadedSound)
			ifTrue: [n, '(out)']
			ifFalse: [n]].
	names add: 'clink'.
	names add: 'edit ', (instrumentSelector at: channelIndex) contents.
	^ names asArray
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
invokeMenu
	"Invoke a menu of additonal commands."

	| aMenu |
	aMenu := CustomMenu new.
	aMenu add: 'add channel' translated action: #addChannel.
	aMenu add: 'reload instruments' translated target: AbstractSound selector: #updateScorePlayers.
	midiSynth isOn ifFalse: [
		aMenu add: 'set MIDI port' translated action: #setMIDIPort.
		midiSynth midiPort
			ifNotNil: [aMenu add: 'close MIDI port' translated action: #closeMIDIPort]].	
	aMenu invokeOn: self defaultSelection: nil.
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
makeControls

	| bb r reverbSwitch onOffSwitch |
	bb := SimpleButtonMorph new
		target: self;
		borderColor: #raised;
		borderWidth: 2;
		color: color.
	r := AlignmentMorph newRow.
	r color: bb color; borderWidth: 0; layoutInset: 0.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	r addMorphBack: (
		bb label: '<>';
			actWhen: #buttonDown;
			actionSelector: #invokeMenu).
	onOffSwitch := SimpleSwitchMorph new
		offColor: color;
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		borderWidth: 2;
		label: 'On';
		actionSelector: #toggleOnOff;
		target: self;
		setSwitchState: false.
	r addMorphBack: onOffSwitch.
	reverbSwitch := SimpleSwitchMorph new
		offColor: color;
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		borderWidth: 2;
		label: 'Reverb Disable';
		actionSelector: #disableReverb:;
		target: self;
		setSwitchState: SoundPlayer isReverbOn not.
	r addMorphBack: reverbSwitch.
	^ r
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'ar 11/9/2000 21:19'!
makeRow

	^ AlignmentMorph newRow
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter;
		hResizing: #spaceFill;
		vResizing: #shrinkWrap
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
panAndVolControlsFor: channelIndex

	| volSlider panSlider c r middleLine |
	volSlider := SimpleSliderMorph new
		color: color;
		extent: 101@2;
		target: midiSynth;
		arguments: (Array with: channelIndex);
		actionSelector: #volumeForChannel:put:;
		minVal: 0.0;
		maxVal: 1.0;
		adjustToValue: (midiSynth volumeForChannel: channelIndex).
	panSlider := SimpleSliderMorph new
		color: color;
		extent: 101@2;
		target: midiSynth;
		arguments: (Array with: channelIndex);
		actionSelector: #panForChannel:put:;
		minVal: 0.0;
		maxVal: 1.0;		
		adjustToValue: (midiSynth panForChannel: channelIndex).
	c := AlignmentMorph newColumn
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #topCenter;
		hResizing: #spaceFill;
		vResizing: #shrinkWrap.
	middleLine := Morph new  "center indicator for pan slider"
		color: (Color r: 0.4 g: 0.4 b: 0.4);
		extent: 1@(panSlider height - 4);
		position: panSlider center x@(panSlider top + 2).
	panSlider addMorphBack: middleLine.
	r := self makeRow.
	r addMorphBack: (StringMorph contents: '0').
	r addMorphBack: volSlider.
	r addMorphBack: (StringMorph contents: '10').
	c addMorphBack: r.
	r := self makeRow.
	r addMorphBack: (StringMorph contents: 'L').
	r addMorphBack: panSlider.
	r addMorphBack: (StringMorph contents: 'R').
	c addMorphBack: r.
	^ c
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
setMIDIPort

	| portNum |
	portNum := SimpleMIDIPort outputPortNumFromUser.
	portNum ifNil: [^ self].
	midiPortNumber := portNum.
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
toggleOnOff

	midiSynth isOn
		ifTrue: [
			midiSynth stopMIDITracking]
		ifFalse: [
			midiPortNumber ifNil: [self setMIDIPort].
			midiPortNumber ifNil: [midiPortNumber := 0].
			midiSynth midiPort: (SimpleMIDIPort openOnPortNumber: midiPortNumber).
			midiSynth startMIDITracking].
! !

!MidiInputMorph methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:41'!
updateInstrumentsFromLibraryExcept: soundsBeingEdited
	"The instrument library has been modified. Update my instruments with the new versions from the library. Use a single instrument prototype for all parts with the same name; this allows the envelope editor to edit all the parts by changing a single sound prototype."

	"soundsBeingEdited is a collection of sounds being edited (by an EnvelopeEditor).  If any of my instruments share one of these, then they will be left alone so as not to disturb that dynamic linkage."

	| unloadPostfix myInstruments name displaysAsUnloaded isUnloaded |
	unloadPostfix := '(out)'.
	myInstruments := Dictionary new.
	1 to: instrumentSelector size do: [:i |
		name := (instrumentSelector at: i) contents.
		displaysAsUnloaded := name endsWith: unloadPostfix.
		displaysAsUnloaded ifTrue: [
			name := name copyFrom: 1 to: name size - unloadPostfix size].
		(myInstruments includesKey: name) ifFalse: [
			myInstruments at: name put:
				(name = 'clink'
					ifTrue: [
						(SampledSound
							samples: SampledSound coffeeCupClink
							samplingRate: 11025) copy]
					ifFalse: [
						(AbstractSound
							soundNamed: name
							ifAbsent: [
								(instrumentSelector at: i) contentsClipped: 'default'.
								FMSound default]) copy])].
		(soundsBeingEdited includes: (midiSynth instrumentForChannel: i)) ifFalse:
			["Do not update any instrument that is currently being edited"
			midiSynth instrumentForChannel: i put: (myInstruments at: name)].

		"update loaded/unloaded status in instrumentSelector if necessary"
		isUnloaded := (myInstruments at: name) isKindOf: UnloadedSound.
		(displaysAsUnloaded and: [isUnloaded not])
			ifTrue: [(instrumentSelector at: i) contentsClipped: name].
		(displaysAsUnloaded not and: [isUnloaded])
			ifTrue: [(instrumentSelector at: i) contentsClipped: name, unloadPostfix]].
! !

AlignmentMorph subclass: #RecordingControlsMorph
	instanceVariableNames: 'recorder recordingStatusLight recordMeter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!RecordingControlsMorph class methodsFor: 'parts bin' stamp: 'sw 8/2/2001 18:18'!
descriptionForPartsBin
	^ self partName:	'SoundRecorder'
		categories:		#('Multimedia')
		documentation:	'A device for making sound recordings.'! !

!RecordingControlsMorph class methodsFor: 'initialization' stamp: 'asm 4/11/2003 10:54'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: #(RecordingControlsMorph	authoringPrototype	'Sound' 	'A device for making sound recordings.')
						forFlapNamed: 'PlugIn Supplies'.
						cl registerQuad: #(RecordingControlsMorph	authoringPrototype	'Sound' 	'A device for making sound recordings.')
						forFlapNamed: 'Widgets'.]! !

!RecordingControlsMorph methodsFor: 'initialization' stamp: 'stephane.ducasse 11/8/2008 15:10'!
addButtonRows

	| r |
	r := AlignmentMorph newRow vResizing: #shrinkWrap.
	r addMorphBack: (self buttonName: 'Morph' translated action: #makeSoundMorph).
	r addMorphBack: (Morph new extent: 4@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Trim' translated action: #trim).
	r addMorphBack: (Morph new extent: 4@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Show' translated action: #show).
	self addMorphBack: r.

	r := AlignmentMorph newRow vResizing: #shrinkWrap.
	r addMorphBack: (self buttonName: 'Record' translated action: #record).
	r addMorphBack: (Morph new extent: 4@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Stop' translated action: #stop).
	r addMorphBack: (Morph new extent: 4@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Play' translated action: #playback).
	r addMorphBack: self makeStatusLight.
	self addMorphBack: r.
! !

!RecordingControlsMorph methodsFor: 'other' stamp: 'ar 11/9/2000 21:21'!
addRecordLevelSlider

	| levelSlider r |
	levelSlider := SimpleSliderMorph new
		color: color;
		extent: 100@2;
		target: recorder;
		actionSelector: #recordLevel:;
		adjustToValue: recorder recordLevel.
	r := AlignmentMorph newRow
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter;
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: '0 ').
	r addMorphBack: levelSlider.
	r addMorphBack: (StringMorph contents: ' 10').
	self addMorphBack: r.
! !

!RecordingControlsMorph methodsFor: 'other' stamp: 'jhm 10/15/97 14:30'!
buttonName: aString action: aSymbol

	^ SimpleButtonMorph new
		target: self;
		label: aString;
		actionSelector: aSymbol
! !

!RecordingControlsMorph methodsFor: 'button commands' stamp: 'stephane.ducasse 11/8/2008 15:11'!
done

	recorder stopRecording.
	self delete.
! !

!RecordingControlsMorph methodsFor: 'initialization' stamp: 'ar 11/9/2000 21:58'!
initialize

	| r |
	super initialize.
	self hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	borderWidth := 2.
	self listDirection: #topToBottom.
	recorder := SoundRecorder new.
	self addButtonRows.
	self addRecordLevelSlider.

	r := AlignmentMorph newRow vResizing: #shrinkWrap.
	r addMorphBack: self makeRecordMeter.
	self addMorphBack: r.
	self extent: 10@10.  "make minimum size"
! !

!RecordingControlsMorph methodsFor: 'other' stamp: 'jm 8/24/97 21:31'!
makeRecordMeter

	| outerBox |
	outerBox := Morph new extent: 102@18; color: Color gray.
	recordMeter := Morph new extent: 1@16; color: Color yellow.
	recordMeter position: outerBox topLeft + (1@1).
	outerBox addMorph: recordMeter.
	^ outerBox
! !

!RecordingControlsMorph methodsFor: 'button commands' stamp: 'sw 6/10/2003 12:59'!
makeSoundMorph

	| m |
	recorder verifyExistenceOfRecordedSound ifFalse: [^ self].

	recorder pause.
	m := SoundEventMorph new sound: recorder recordedSound.
	self world firstHand attachMorph: m.
! !

!RecordingControlsMorph methodsFor: 'other' stamp: 'jhm 10/15/97 15:06'!
makeStatusLight

	recordingStatusLight := Morph new extent: 18@18.
	recordingStatusLight color: Color transparent.
	^ recordingStatusLight
! !

!RecordingControlsMorph methodsFor: 'button commands' stamp: 'sw 6/10/2003 12:59'!
playback
	"The user hit the playback button"

	recorder verifyExistenceOfRecordedSound ifFalse: [^ self].
	recorder pause.
	recorder playback.
! !

!RecordingControlsMorph methodsFor: 'button commands' stamp: 'jhm 10/15/97 14:30'!
record

	recorder clearRecordedSound.
	recorder resumeRecording.
! !

!RecordingControlsMorph methodsFor: 'accessing' stamp: 'jm 9/13/97 17:20'!
recorder

	^ recorder
! !

!RecordingControlsMorph methodsFor: 'button commands' stamp: 'sw 6/10/2003 12:59'!
show
	"Show my samples in a WaveEditor."

	| ed w |
	recorder verifyExistenceOfRecordedSound ifFalse: [^ self].
	recorder pause.
	ed := WaveEditor new.
	ed data: recorder condensedSamples.
	ed samplingRate: recorder samplingRate.
	w := self world.
	w activeHand
		ifNil: [w addMorph: ed]
		ifNotNil: [w activeHand attachMorph: ed].

! !

!RecordingControlsMorph methodsFor: 'stepping and presenter' stamp: 'tk 6/24/1999 11:41'!
startStepping
	"Make the level meter active when dropped into the world. Do nothing if already recording. Note that this will cause other recorders to stop recording..."

	super startStepping.
	recorder isPaused ifTrue: [
		SoundRecorder allSubInstancesDo: [:r | r stopRecording].  "stop all other sound recorders"
		recorder pause].  "meter is updated while paused"
! !

!RecordingControlsMorph methodsFor: 'stepping and presenter' stamp: 'jm 10/17/97 15:00'!
step

	recorder isPaused
		ifTrue: [recordingStatusLight color: Color transparent]
		ifFalse: [recordingStatusLight color: Color red].
	recordMeter extent: (recorder meterLevel + 1) @ recordMeter height.
! !

!RecordingControlsMorph methodsFor: 'testing' stamp: 'jm 8/24/97 21:00'!
stepTime

	^ 50
! !

!RecordingControlsMorph methodsFor: 'stepping and presenter' stamp: 'jhm 10/15/97 14:30'!
stop

	recorder pause.
! !

!RecordingControlsMorph methodsFor: 'stepping and presenter' stamp: 'jm 10/17/97 15:13'!
stopStepping
	"Turn off recording."

	super stopStepping.
	recorder stopRecording.
! !

!RecordingControlsMorph methodsFor: 'button commands' stamp: 'sw 6/10/2003 12:59'!
trim
	"Show my samples in a GraphMorph."
	
	recorder verifyExistenceOfRecordedSound ifFalse: [^ self].
	recorder pause.
	recorder trim: 1400 normalizedVolume: 80.0.
! !

AlignmentMorph subclass: #ScorePlayerMorph
	instanceVariableNames: 'scorePlayer trackInstNames instrumentSelector scrollSlider'
	classVariableNames: 'LastMIDIPort'
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!ScorePlayerMorph commentStamp: '<historical>' prior: 0!
A ScorePlayerMorph mediates between a score such as a MIDIScore, a PianoRollScoreMorph, and the actual SoundPlayer synthesizer.

It provides control over volume, tempo, instrumentation, and location in the score.!

!ScorePlayerMorph class methodsFor: 'parts bin' stamp: 'sw 10/20/2001 17:17'!
descriptionForPartsBin
	^ self partName: 	'ScorePlayer'
		categories:		#('Multimedia')
		documentation:	' Mediates between a score such as a MIDIScore, a PianoRollScoreMorph, and the actual SoundPlayer synthesizer'! !

!ScorePlayerMorph class methodsFor: 'filein/out' stamp: 'sd 2/6/2002 21:37'!
fileReaderServicesForFile: fullName suffix: suffix

	^(suffix = 'mid') | (suffix = '*') 
		ifTrue: [ self services]
		ifFalse: [#()]
! !

!ScorePlayerMorph class methodsFor: 'initialization' stamp: 'GabrielOmarCotelli 6/4/2009 20:32'!
initialize

	FileServices registerFileReader: self! !

!ScorePlayerMorph class methodsFor: 'system hookup' stamp: 'stephaneducasse 2/4/2006 20:41'!
onMIDIFileNamed: fileName
	"Return a ScorePlayerMorph on the score from the MIDI file of the given name."

	| score player |
	score := MIDIFileReader scoreFromFileNamed: fileName	.
	player := ScorePlayer onScore: score.
	^ self new onScorePlayer: player title: fileName
! !

!ScorePlayerMorph class methodsFor: 'system hookup' stamp: 'stephaneducasse 2/4/2006 20:41'!
openOn: aScore title: aString

	| player |
	player := ScorePlayer onScore: aScore.
	(self new onScorePlayer: player title: aString) openInWorld.
! !

!ScorePlayerMorph class methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
playMidiFile: fullName
	"Play a MIDI file."
 
	| f score |
	Smalltalk at: #MIDIFileReader ifPresent: [:midiReader |
			f := (FileStream oldFileNamed: fullName) binary.
			score := (midiReader new readMIDIFrom: f) asScore.
			f close.
			self openOn: score title: (FileDirectory localNameFor: fullName)]
! !

!ScorePlayerMorph class methodsFor: 'initialization' stamp: 'sw 2/17/2002 02:45'!
servicePlayMidiFile
	"Answer a service for opening player on a midi file"

	^ SimpleServiceEntry 
		provider: self 
		label: 'open in midi player'
		selector: #playMidiFile:
		description: 'open the midi-player tool on this file'
		buttonLabel: 'open'! !

!ScorePlayerMorph class methodsFor: 'filein/out' stamp: 'sd 2/1/2002 22:21'!
services

	^ Array with: self servicePlayMidiFile

	! !

!ScorePlayerMorph class methodsFor: 'initialize-release' stamp: 'GabrielOmarCotelli 6/4/2009 20:33'!
unload

	FileServices unregisterFileReader: self ! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'gm 2/28/2003 00:28'!
atTrack: trackIndex from: aPopUpChoice selectInstrument: selection 
	| oldSnd name snd |
	oldSnd := scorePlayer instrumentForTrack: trackIndex.
	(selection beginsWith: 'edit ') 
		ifTrue: 
			[name := selection copyFrom: 6 to: selection size.
			aPopUpChoice contentsClipped: name.
			(oldSnd isKindOf: FMSound) | (oldSnd isKindOf: LoopedSampledSound) 
				ifTrue: [EnvelopeEditorMorph openOn: oldSnd title: name].
			(oldSnd isKindOf: SampledInstrument) 
				ifTrue: [EnvelopeEditorMorph openOn: oldSnd allNotes first title: name].
			^self].
	snd := nil.
	1 to: instrumentSelector size
		do: 
			[:i | 
			(trackIndex ~= i and: [selection = (instrumentSelector at: i) contents]) 
				ifTrue: [snd := scorePlayer instrumentForTrack: i]].	"use existing instrument prototype"
	snd ifNil: 
			[snd := (selection = 'clink' 
				ifTrue: 
					[(SampledSound samples: SampledSound coffeeCupClink
								samplingRate: 11025) ]
				ifFalse: [(AbstractSound soundNamed: selection)]) copy].
	scorePlayer instrumentForTrack: trackIndex put: snd.
	(instrumentSelector at: trackIndex) contentsClipped: selection! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
closeMIDIPort

	scorePlayer closeMIDIPort.
	LastMIDIPort := nil.
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:39'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:30'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color veryLightGray! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:25'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self listDirection: #topToBottom;
		 wrapCentering: #center;
		 cellPositioning: #topCenter;
		 hResizing: #shrinkWrap;
		 vResizing: #shrinkWrap;
		 layoutInset: 3;
		 onScorePlayer: ScorePlayer new initialize title: ' ';
		 extent: 20 @ 20 ! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
instrumentChoicesForTrack: trackIndex
	| names inst |

	names := AbstractSound soundNames asOrderedCollection.
	names := names collect: [:n |
		inst := AbstractSound soundNamed: n.
		(inst isKindOf: UnloadedSound)
			ifTrue: [n, '(out)']
			ifFalse: [n]].
	names add: 'clink'.
	names add: 'edit ', (instrumentSelector at: trackIndex) contents.
	^ names asArray
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
invokeMenu
	"Invoke a menu of additonal functions for this ScorePlayer."

	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu add: 'open a MIDI file' translated action: #openMIDIFile.
	aMenu addList: {
		#-.
		{'save as AIFF file' translated.	#saveAsAIFF}.
		{'save as WAV file' translated.		#saveAsWAV}.
		{'save as Sun AU file' translated.	#saveAsSunAudio}.
		#-}.
	aMenu add: 'reload instruments' translated target: AbstractSound selector: #updateScorePlayers.
	aMenu addLine.
	scorePlayer midiPort
		ifNil: [
			aMenu add: 'play via MIDI' translated action: #openMIDIPort]
		ifNotNil: [
			aMenu add: 'play via built in synth' translated action: #closeMIDIPort.
			aMenu add: 'new MIDI controller' translated action: #makeMIDIController:].
	aMenu addLine.
	aMenu add: 'make a pause marker' translated action: #makeAPauseEvent:.

	aMenu popUpInWorld: self world.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
makeAPauseEvent: evt

	| newWidget |

	newWidget := AlignmentMorph newRow.
	newWidget 
		color: Color orange; 
		borderWidth: 0; 
		layoutInset: 0;
		hResizing: #shrinkWrap; 
		vResizing: #shrinkWrap; 
		extent: 5@5;
		addMorph: (StringMorph contents: '[pause]' translated) lock;
		addMouseUpActionWith: (
			MessageSend receiver: self selector: #showResumeButtonInTheWorld
		).

	evt hand attachMorph: newWidget.! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
makeControls

	| bb r reverbSwitch repeatSwitch |
	r := AlignmentMorph newRow.
	r color: color; borderWidth: 0; layoutInset: 0.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	bb := SimpleButtonMorph new target: self; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: '<>'; actWhen: #buttonDown;
												actionSelector: #invokeMenu).
	bb := SimpleButtonMorph new target: self; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Piano Roll' translated;		actionSelector: #makePianoRoll).
	bb := SimpleButtonMorph new target: self; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Rewind' translated;		actionSelector: #rewind).
	bb := SimpleButtonMorph new target: scorePlayer; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Play' translated;			actionSelector: #resumePlaying).
	bb := SimpleButtonMorph new target: scorePlayer; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Pause' translated;			actionSelector: #pause).
	reverbSwitch := SimpleSwitchMorph new
		offColor: color;
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		borderWidth: 2;
		label: 'Reverb Disable' translated;
		actionSelector: #disableReverb:;
		target: scorePlayer;
		setSwitchState: SoundPlayer isReverbOn not.
	r addMorphBack: reverbSwitch.
	scorePlayer ifNotNil:
		[repeatSwitch := SimpleSwitchMorph new
			offColor: color;
			onColor: (Color r: 1.0 g: 0.6 b: 0.6);
			borderWidth: 2;
			label: 'Repeat' translated;
			actionSelector: #repeat:;
			target: scorePlayer;
			setSwitchState: scorePlayer repeat.
		r addMorphBack: repeatSwitch].
	^ r
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'jm 9/28/1998 23:05'!
makeMIDIController: evt

	self world activeHand attachMorph:
		(MIDIControllerMorph new midiPort: scorePlayer midiPort).
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
makePianoRoll
	"Create a piano roll viewer for this score player."

	| pianoRoll hand |
	pianoRoll := PianoRollScoreMorph new on: scorePlayer.
	hand := self world activeHand.
	hand ifNil: [self world addMorph: pianoRoll]
		ifNotNil: [hand attachMorph: pianoRoll.
				hand lastEvent shiftPressed ifTrue:
					["Special case for NOBM demo"
					pianoRoll contractTime; contractTime; enableDragNDrop]].
	pianoRoll startStepping.
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'ar 11/9/2000 21:23'!
makeRow

	^ AlignmentMorph newRow
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter;
		hResizing: #spaceFill;
		vResizing: #shrinkWrap
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
onScorePlayer: aScorePlayer title: scoreName
	| divider col r |
	scorePlayer := aScorePlayer.
	scorePlayer ifNotNil:
		[scorePlayer  reset.
		instrumentSelector := Array new: scorePlayer score tracks size].

	self removeAllMorphs.
	self addMorphBack: self makeControls.
	scorePlayer ifNil: [^ self].

	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: self rateControl;
		addMorphBack: (Morph newBounds: (0@0 extent: 20@0) color: Color transparent);
		addMorphBack: self volumeControl.
	self addMorphBack: r.
	self addMorphBack: self scrollControl.

	col := AlignmentMorph newColumn color: color; layoutInset: 0.
	self addMorphBack: col.
	1 to: scorePlayer trackCount do: [:trackIndex |
		divider := AlignmentMorph new
			extent: 10@1;
			borderWidth: 1;
			layoutInset: 0;
			borderColor: #raised;
			color: color;
			hResizing: #spaceFill;
			vResizing: #rigid.
		col addMorphBack: divider.
		col addMorphBack: (self trackControlsFor: trackIndex)].

	LastMIDIPort ifNotNil: [
		"use the most recently set MIDI port"
		scorePlayer openMIDIPort: LastMIDIPort].
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'adrian_lienhard 7/18/2009 16:00'!
openMIDIFile
	"Open a MIDI score and re-init controls..."
	| score fileName f player |
	fileName := Utilities chooseFileWithSuffixFromList: #('.mid' '.midi')
					withCaption: 'Choose a MIDI file to open' translated.
	(fileName isNil or: [ fileName == #none ])
		ifTrue: [^ self inform: 'No .mid/.midi files found in the Pharo directory' translated].
	f := FileStream readOnlyFileNamed: fileName.
	score := (MIDIFileReader new readMIDIFrom: f binary) asScore.
	f close.
	player := ScorePlayer onScore: score.
	self onScorePlayer: player title: fileName! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
openMIDIPort

	| portNum |
	portNum := SimpleMIDIPort outputPortNumFromUser.
	portNum ifNil: [^ self].
	scorePlayer openMIDIPort: portNum.
	LastMIDIPort := portNum.
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
panAndVolControlsFor: trackIndex

	| volSlider panSlider c r middleLine pianoRollColor |
	pianoRollColor := (Color wheel: scorePlayer score tracks size) at: trackIndex.
	volSlider := SimpleSliderMorph new
		color: color;
		sliderColor: pianoRollColor;
		extent: 101@2;
		target: scorePlayer;
		arguments: (Array with: trackIndex);
		actionSelector: #volumeForTrack:put:;
		minVal: 0.0;
		maxVal: 1.0;
		adjustToValue: (scorePlayer volumeForTrack: trackIndex).
	panSlider := SimpleSliderMorph new
		color: color;
		sliderColor: pianoRollColor;
		extent: 101@2;
		target: scorePlayer;
		arguments: (Array with: trackIndex);
		actionSelector: #panForTrack:put:;
		minVal: 0.0;
		maxVal: 1.0;		
		adjustToValue: (scorePlayer panForTrack: trackIndex).
	c := AlignmentMorph newColumn
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #topCenter;
		hResizing: #spaceFill;
		vResizing: #shrinkWrap.
	middleLine := Morph new  "center indicator for pan slider"
		color: (Color r: 0.4 g: 0.4 b: 0.4);
		extent: 1@(panSlider height - 4);
		position: panSlider center x@(panSlider top + 2).
	panSlider addMorphBack: middleLine.
	r := self makeRow.
	r addMorphBack: (StringMorph contents: '0').
	r addMorphBack: volSlider.
	r addMorphBack: (StringMorph contents: '10').
	c addMorphBack: r.
	r := self makeRow.
	r addMorphBack: (StringMorph contents: 'L' translated).
	r addMorphBack: panSlider.
	r addMorphBack: (StringMorph contents: 'R' translated).
	c addMorphBack: r.
	^ c
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
rateControl

	| rateSlider middleLine r |
	rateSlider := SimpleSliderMorph new
		color: color;
		sliderColor: Color gray;
		extent: 180@2;
		target: self;
		actionSelector: #setLogRate:;
		minVal: -1.0;
		maxVal: 1.0;
		adjustToValue: 0.0.
	middleLine := Morph new  "center indicator for pan slider"
		color: (Color r: 0.4 g: 0.4 b: 0.4);
		extent: 1@(rateSlider height - 4);
		position: rateSlider center x@(rateSlider top + 2).
	rateSlider addMorphBack: middleLine.
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'slow ' translated).
	r addMorphBack: rateSlider.
	r addMorphBack: (StringMorph contents: ' fast' translated).
	^ r
! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'jm 10/12/1998 17:18'!
rewind

	scorePlayer pause; reset.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:48'!
saveAsAIFF
	"Create a stereo AIFF audio file with the result of performing my score."

	| fileName |
	fileName := UIManager default request: 'New file name?' translated.
	fileName isEmpty ifTrue: [^ self].
	(fileName asLowercase endsWith: '.aif') ifFalse: [
		fileName := fileName, '.aif'].

	scorePlayer storeAIFFOnFileNamed: fileName.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:48'!
saveAsSunAudio
	"Create a stereo Sun audio file with the result of performing my score."

	| fileName |
	fileName := UIManager default request: 'New file name?' translated.
	fileName isEmpty ifTrue: [^ self].
	(fileName asLowercase endsWith: '.au') ifFalse: [
		fileName := fileName, '.au'].

	scorePlayer storeSunAudioOnFileNamed: fileName.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:48'!
saveAsWAV
	"Create a stereo WAV audio file with the result of performing my score."

	| fileName |
	fileName := UIManager default request: 'New file name?' translated.
	fileName isEmpty ifTrue: [^ self].
	(fileName asLowercase endsWith: '.wav') ifFalse: [
		fileName := fileName, '.wav'].

	scorePlayer storeWAVOnFileNamed: fileName.
! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'jm 6/1/1998 09:10'!
scorePlayer

	^ scorePlayer
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
scrollControl

	| r |
	scrollSlider := SimpleSliderMorph new
		color: color;
		sliderColor: Color gray;
		extent: 360@2;
		target: scorePlayer;
		actionSelector: #positionInScore:;
		adjustToValue: scorePlayer positionInScore.
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'start ' translated).
	r addMorphBack: scrollSlider.
	r addMorphBack: (StringMorph contents: ' end' translated).
	^ r
! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'jm 1/30/98 15:23'!
setLogRate: logOfRate

	scorePlayer rate: (3.5 raisedTo: logOfRate).
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
showResumeButtonInTheWorld

	| w |

	WorldState addDeferredUIMessage: [
		w := self world.
		w ifNotNil: [
			w addMorphFront:
				(self standaloneResumeButton position: (w right - 100) @ (w top + 10)).
			scorePlayer pause.
			].
	]
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
standaloneResumeButton

	| r |

	r := AlignmentMorph newRow.
	r color: Color red; borderWidth: 0; layoutInset: 6; useRoundedCorners.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	r addMorphBack: (
		SimpleButtonMorph new
			target: [
				scorePlayer resumePlaying.
				r delete
			];
			borderColor: #raised;
			borderWidth: 2;
			color: Color green;
			label: 'Continue' translated;
			actionSelector: #value
	).
	r setBalloonText: 'Continue playing a paused presentation' translated.
	^r


! !

!ScorePlayerMorph methodsFor: 'stepping and presenter' stamp: 'di 6/15/1999 11:48'!
step

	scrollSlider adjustToValue: scorePlayer positionInScore.

! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
trackControlsFor: trackIndex

	| r |
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: (self trackNumAndMuteButtonFor: trackIndex).
	r addMorphBack: (Morph new extent: 10@5; color: color).  "spacer"
	r addMorphBack: (self panAndVolControlsFor: trackIndex).
	^ r
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
trackNumAndMuteButtonFor: trackIndex

	| muteButton instSelector pianoRollColor r |
	muteButton := SimpleSwitchMorph new
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		offColor: color;
		color: color;
		label: 'Mute' translated;
		target: scorePlayer;
		actionSelector: #mutedForTrack:put:;
		arguments: (Array with: trackIndex).
	instSelector := PopUpChoiceMorph new
		extent: 95@14;
		contentsClipped: 'oboe1';
		target: self;
		actionSelector: #atTrack:from:selectInstrument:;
		getItemsSelector: #instrumentChoicesForTrack:;
		getItemsArgs: (Array with: trackIndex).
	instSelector arguments:
		(Array with: trackIndex with: instSelector).
	instrumentSelector at: trackIndex put: instSelector.

	"select track color using same color list as PianoRollScoreMorph"
	pianoRollColor := (Color wheel: scorePlayer score tracks size) at: trackIndex.

	r := self makeRow
		hResizing: #rigid;
		vResizing: #spaceFill;
		extent: 70@10.
	r addMorphBack:
		((StringMorph
			contents: trackIndex printString
			font: (TextStyle default fontOfSize: 24)) color: pianoRollColor).
	trackIndex < 10
		ifTrue: [r addMorphBack: (Morph new color: color; extent: 19@8)]  "spacer"
		ifFalse: [r addMorphBack: (Morph new color: color; extent: 8@8)].  "spacer"
	r addMorphBack:
		(StringMorph new
			extent: 140@14;
			contentsClipped: (scorePlayer infoForTrack: trackIndex)).
	r addMorphBack: (Morph new color: color; extent: 8@8).  "spacer"
	r addMorphBack: instSelector.
	r addMorphBack: (AlignmentMorph newRow color: color).  "spacer"
	r addMorphBack: muteButton.
	^ r
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
updateInstrumentsFromLibraryExcept: soundsBeingEdited
	"The instrument library has been modified. Update my instruments with the new versions from the library. Use a single instrument prototype for all parts with the same name; this allows the envelope editor to edit all the parts by changing a single sound prototype."

	"soundsBeingEdited is a collection of sounds being edited (by an EnvelopeEditor).  If any of my instruments share one of these, then they will be left alone so as not to disturb that dynamic linkage."

	| unloadPostfix myInstruments name displaysAsUnloaded isUnloaded |
	unloadPostfix := '(out)'.
	myInstruments := Dictionary new.
	1 to: instrumentSelector size do: [:i |
		name := (instrumentSelector at: i) contents.
		displaysAsUnloaded := name endsWith: unloadPostfix.
		displaysAsUnloaded ifTrue: [
			name := name copyFrom: 1 to: name size - unloadPostfix size].
		(myInstruments includesKey: name) ifFalse: [
			myInstruments at: name put:
				(name = 'clink'
					ifTrue: [
						(SampledSound
							samples: SampledSound coffeeCupClink
							samplingRate: 11025) copy]
					ifFalse: [
						(AbstractSound
							soundNamed: name
							ifAbsent: [
								(instrumentSelector at: i) contentsClipped: 'default'.
								FMSound default]) copy])].
		(soundsBeingEdited includes: (scorePlayer instrumentForTrack: i)) ifFalse:
			["Do not update any instrument that is currently being edited"
			scorePlayer instrumentForTrack: i put: (myInstruments at: name)].

		"update loaded/unloaded status in instrumentSelector if necessary"
		isUnloaded := (myInstruments at: name) isKindOf: UnloadedSound.
		(displaysAsUnloaded and: [isUnloaded not])
			ifTrue: [(instrumentSelector at: i) contentsClipped: name].
		(displaysAsUnloaded not and: [isUnloaded])
			ifTrue: [(instrumentSelector at: i) contentsClipped: name, unloadPostfix]].
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
volumeControl

	| volumeSlider r |
	volumeSlider := SimpleSliderMorph new
		color: color;
		sliderColor: Color gray;
		extent: 80@2;
		target: scorePlayer;
		actionSelector: #overallVolume:;
		adjustToValue: scorePlayer overallVolume.
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'soft  ' translated).
	r addMorphBack: volumeSlider.
	r addMorphBack: (StringMorph contents: ' loud' translated).
	^ r
! !

!ScorePlayerMorph methodsFor: 'rounding' stamp: 'di 10/22/2000 23:42'!
wantsRoundedCorners
	^ Preferences roundedWindowCorners or: [super wantsRoundedCorners]! !

AlignmentMorph subclass: #SoundDemoMorph
	instanceVariableNames: 'soundColumn'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SoundDemoMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:40'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!SoundDemoMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color lightGray! !

!SoundDemoMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:04'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self listDirection: #topToBottom;
		 wrapCentering: #center;
		 cellPositioning: #topCenter;
		 hResizing: #spaceFill;
		 vResizing: #spaceFill;
		 layoutInset: 3;
		 addMorph: self makeControls;
	initializeSoundColumn.
	self extent: 118 @ 150! !

!SoundDemoMorph methodsFor: 'initialization' stamp: 'jam 3/9/2003 17:52'!
initializeSoundColumn
"initialize the receiver's soundColumn"
	soundColumn := AlignmentMorph newColumn.
	soundColumn enableDragNDrop.
	self addMorphBack: soundColumn! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'tk 2/19/2001 17:49'!
makeControls

	| bb r cc |
	cc := Color black.
	r := AlignmentMorph newRow.
	r color: cc; borderWidth: 0; layoutInset: 0.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	r addMorphBack: (bb label: 'V1';			actionSelector: #playV1).
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	r addMorphBack: (bb label: 'V2';			actionSelector: #playV2).
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	r addMorphBack: (bb label: 'V3';			actionSelector: #playV3).
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	r addMorphBack: (bb label: 'All';			actionSelector: #playAll).
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	r addMorphBack: (bb label: 'Stop';		actionSelector: #stopSound).
	^ r
! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'dgd 2/22/2003 13:34'!
playAll
	| snd |
	soundColumn submorphs isEmpty
		ifTrue: [^ self].
	self setTimbreFromTile: soundColumn submorphs first.
	snd := SampledSound bachFugueVoice1On: SampledSound new.
	soundColumn submorphs size >= 2
		ifTrue: [""self setTimbreFromTile: soundColumn submorphs second.
			snd := snd
						+ (AbstractSound bachFugueVoice2On: SampledSound new)].
	soundColumn submorphs size >= 3
		ifTrue: [""self setTimbreFromTile: soundColumn submorphs third.
			snd := snd
						+ (AbstractSound bachFugueVoice3On: SampledSound new)].
	snd play! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'dgd 2/22/2003 13:35'!
playV1
	soundColumn submorphs isEmpty
		ifTrue: [^ self].
	self
		setTimbreFromTile: (soundColumn submorphs first).
	(SampledSound bachFugueVoice1On: SampledSound new) play! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'dgd 2/22/2003 13:35'!
playV2
	soundColumn submorphs size < 2
		ifTrue: [^ self].
	self
		setTimbreFromTile: (soundColumn submorphs second).
	(SampledSound bachFugueVoice2On: SampledSound new) playSilentlyUntil: 4.8;
		 resumePlaying! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'dgd 2/22/2003 13:35'!
playV3
	soundColumn submorphs size < 3
		ifTrue: [^ self].
	self
		setTimbreFromTile: (soundColumn submorphs third).
	(AbstractSound bachFugueVoice3On: SampledSound new) playSilentlyUntil: 14.4;
		 resumePlaying! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:09'!
setTimbreFromTile: aSoundTile

	SampledSound defaultSampleTable: aSoundTile sound samples.
	SampledSound nominalSamplePitch: 400.
! !

!SoundDemoMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:09'!
stopSound

	SoundPlayer shutDown.
! !

AlignmentMorph subclass: #SpectrumAnalyzerMorph
	instanceVariableNames: 'soundInput statusLight levelMeter graphMorph sonogramMorph fft displayType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SpectrumAnalyzerMorph commentStamp: '<historical>' prior: 0!
I am a tool for analyzing sound data from a microphone, CD, or other input source in real time. I have several display modes:

	signal		snapshots of the raw signal data as it arrives
	spectrum	frequency spectrum of the signal data as it arrives
	sonogram	scrolling plot of the frequency spectrum over time,
			      where the vertical axis is frequency, the horizontal
				  axis is time, and amount of energy at a given
				  frequency is shown as a grayscale value with
				  larger values being darker

To use this tool, be sure that you have selected the proper sound source using you host OS facilities. Set the desired sampling rate and FFT size (try 22050 samples/sec and an FFT size of 512) then click on the 'start' button. Use the slider to adjust the level so that the yellow level indicator peaks somewhere between the middle and the right edge at the maximum signal level.

Note that if the level meter peaks hit the right edge, you will get 'clipping', which creates a bunch of spurious high frequency noise in the frequency spectrum. If the display is set to 'signal' mode, you can actually see the tops and bottoms of the waveform being cut off when clipping occurs.

Many machines may not be able to perform spectrum analysis in real time, especially at higher sampling rates and larger FFT sizes. In both 'signal' and 'spectrum' modes, this tool will skip data to try to keep up with real time. However, in 'sonogram' mode it always processes all the data, even if it falls behind. This allows you to get a complete sonogram without dropouts even on a slower machine. However, as the sonogram display falls behind there will be a larger and larger time lag between when a sound is input and when it appears on the display.

The smaller the FFT size, the less frequency resolution you get. The lower the sampling rate, the less total frequency range you get. For an FFT size of N and a sampling rate of R, each of the N/2 'bins' of the frequency spectrum has a frequency resolution of R / N. For example, at a sampleing rate of 22050 samples/second, the total frequency range is 0 to 11025 Hz and an FFT of size 256 would divide this range into 128 bins (the output of an FFT of size N has N/2 bins), each of which covers a frequency band about 86 Hz wide.

To increase time resolution, increase the sampling rate and decrease the FFT size.
!

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'dgd 9/19/2003 13:30'!
addButtonRow

	| r |
	r := AlignmentMorph newRow vResizing: #shrinkWrap.
	r addMorphBack: (self buttonName: 'Menu' translated action: #invokeMenu).
	r addMorphBack: (Morph new extent: 4@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Start' translated action: #start).
	r addMorphBack: (Morph new extent: 4@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Stop' translated action: #stop).
	r addMorphBack: (Morph new extent: 12@1; color: Color transparent).
	r addMorphBack: self makeStatusLight.
	self addMorphBack: r.
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'ar 11/9/2000 21:23'!
addLevelSlider

	| levelSlider r |
	levelSlider := SimpleSliderMorph new
		color: color;
		extent: 100@2;
		target: soundInput;
		actionSelector: #recordLevel:;
		adjustToValue: soundInput recordLevel.
	r := AlignmentMorph newRow
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter;
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: '0 ').
	r addMorphBack: levelSlider.
	r addMorphBack: (StringMorph contents: ' 10').
	self addMorphBack: r.
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jhm 10/15/97 14:30'!
buttonName: aString action: aSymbol

	^ SimpleButtonMorph new
		target: self;
		label: aString;
		actionSelector: aSymbol
! !

!SpectrumAnalyzerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:52'!
defaultBorderWidth
"answer the default border width for the receiver"
	^ 2! !

!SpectrumAnalyzerMorph methodsFor: 'submorphs-add/remove' stamp: 'jm 9/6/1999 14:40'!
delete
	"Turn off recording when this morph is deleted."

	super delete.
	soundInput stopRecording.
! !

!SpectrumAnalyzerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:52'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	self listDirection: #topToBottom.
	soundInput := SoundInputStream new samplingRate: 22050.
	fft := FFT new: 512.
	displayType := 'sonogram'.
	self hResizing: #shrinkWrap.
	self vResizing: #shrinkWrap.
	self addButtonRow.
	self addLevelSlider.
	self addMorphBack: self makeLevelMeter.
	self addMorphBack: (Morph new extent: 10 @ 10;
			 color: Color transparent).
	"spacer"
	self resetDisplay! !

!SpectrumAnalyzerMorph methodsFor: 'menu and buttons' stamp: 'dgd 9/19/2003 13:25'!
invokeMenu
	"Invoke the settings menu."

	| aMenu |
	aMenu := CustomMenu new.
	aMenu addList:	{
		{'set sampling rate' translated.		#setSamplingRate}.
		{'set FFT size' translated.			#setFFTSize}.
		{'set display type' translated.		#setDisplayType}}.
	aMenu invokeOn: self defaultSelection: nil.
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jm 9/8/1999 13:27'!
makeLevelMeter

	| outerBox |
	outerBox := RectangleMorph new extent: 125@14; color: Color lightGray.
	levelMeter := Morph new extent: 2@10; color: Color yellow.
	levelMeter position: outerBox topLeft + (2@2).
	outerBox addMorph: levelMeter.
	^ outerBox
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'dgd 9/19/2003 13:31'!
makeStatusLight

	| s |
	statusLight := RectangleMorph new extent: 24@19.
	statusLight color: Color gray.
	s := StringMorph contents: 'On' translated.
	s position: statusLight center - (s extent // 2).
	statusLight addMorph: s.
	^ statusLight
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'gm 2/28/2003 00:11'!
processBuffer: buf 
	"Analyze one buffer of data."

	| data |
	data := displayType = 'signal' 
		ifTrue: [buf]
		ifFalse: [fft transformDataFrom: buf startingAt: 1].
	graphMorph ifNotNil: 
			[graphMorph
				data: data;
				changed].
	sonogramMorph ifNotNil: 
			[data := data collect: [:v | v sqrt].	"square root compresses dynamic range"
			data /= 400.0.
			sonogramMorph plotColumn: (data copyFrom: 1 to: data size // 1)]! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jm 9/8/1999 12:49'!
removeAllDisplays
	"Remove all currently showing displays."

	sonogramMorph ifNotNil: [sonogramMorph delete].
	graphMorph ifNotNil: [graphMorph delete].
	sonogramMorph := graphMorph := nil.
! !

!SpectrumAnalyzerMorph methodsFor: 'menu and buttons' stamp: 'jm 9/8/1999 12:52'!
resetDisplay
	"Recreate my display after changing some parameter such as FFT size."

	displayType = 'signal' ifTrue: [self showSignal].
	displayType = 'spectrum' ifTrue: [self showSpectrum].
	displayType = 'sonogram' ifTrue: [self showSonogram].
! !

!SpectrumAnalyzerMorph methodsFor: 'menu and buttons' stamp: 'dgd 9/19/2003 13:29'!
setDisplayType
	"Set the display type."

	| aMenu choice on |
	aMenu := CustomMenu new title: ('display type (currently {1})' translated format:{displayType}).
	aMenu addList:	{
		{'signal' translated.	'signal'}.
		{'spectrum' translated.	'spectrum'}.
		{'sonogram' translated.	'sonogram'}}.
	choice := aMenu startUp.
	choice ifNil: [^ self].

	on := soundInput isRecording.
	self stop.
	displayType := choice.
	self resetDisplay.
	on ifTrue: [self start].

! !

!SpectrumAnalyzerMorph methodsFor: 'menu and buttons' stamp: 'dgd 9/19/2003 13:27'!
setFFTSize
	"Set the size of the FFT used for frequency analysis."

	| aMenu sz on |
	aMenu := CustomMenu new title: ('FFT size (currently {1})' translated format:{fft n}).
	((7 to: 10) collect: [:n | 2 raisedTo: n]) do:[:r | aMenu add: r printString action: r].
	sz := aMenu startUp.
	sz ifNil: [^ self].
	on := soundInput isRecording.
	self stop.
	fft := FFT new: sz.
	self resetDisplay.
	on ifTrue: [self start].
! !

!SpectrumAnalyzerMorph methodsFor: 'menu and buttons' stamp: 'dgd 9/19/2003 13:26'!
setSamplingRate
	"Set the sampling rate to be used for incoming sound data."

	| aMenu rate on |
	aMenu := CustomMenu new title:
		('Sampling rate (currently {1})' translated format:{soundInput samplingRate}).
	#(11025 22050 44100) do:[:r | aMenu add: r printString action: r].
	rate := aMenu startUp.
	rate ifNil: [^ self].
	on := soundInput isRecording.
	self stop.
	soundInput samplingRate: rate.
	self resetDisplay.
	on ifTrue: [self start].

! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jm 9/8/1999 19:56'!
showSignal
	"Display the actual signal waveform."

	displayType := 'signal'.
	self removeAllDisplays.
	graphMorph := GraphMorph new.
	graphMorph extent: (400 + (2 * graphMorph borderWidth))@128.
	graphMorph data: (Array new: 100 withAll: 0).
	graphMorph color: (Color r: 0.8 g: 1.0 b: 1.0).
	self addMorphBack: graphMorph.
	self extent: 10@10.  "shrink to minimum size"
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jm 9/8/1999 19:43'!
showSonogram
	"Display a sonogram showing the frequency spectrum versus time."

	| zeros h w |
	displayType := 'sonogram'.
	self removeAllDisplays.
	h := fft n // 2.
	h := h min: 512 max: 64.
	w := 400.
	sonogramMorph :=
		Sonogram new
			extent: w@h
			minVal: 0.0
			maxVal: 1.0
			scrollDelta: w.
	zeros := Array new: sonogramMorph height withAll: 0.
	sonogramMorph width timesRepeat: [sonogramMorph plotColumn: zeros].
	self addMorphBack: sonogramMorph.
	self extent: 10@10.  "shrink to minimum size"
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jm 9/8/1999 15:10'!
showSpectrum
	"Display the frequency spectrum."

	displayType := 'spectrum'.
	self removeAllDisplays.
	graphMorph := GraphMorph new.
	graphMorph extent: ((fft n // 2) + (2 * graphMorph borderWidth))@128.
	graphMorph data: (Array new: fft n // 2 withAll: 0).
	self addMorphBack: graphMorph.
	self extent: 10@10.  "shrink to minimum size"
! !

!SpectrumAnalyzerMorph methodsFor: 'stepping and presenter' stamp: 'jm 9/8/1999 15:12'!
start
	"Start displaying sound data."

	displayType = 'signal'
		ifTrue: [soundInput bufferSize: graphMorph width - (2 * graphMorph borderWidth)]
		ifFalse: [soundInput bufferSize: fft n].
	soundInput startRecording.
! !

!SpectrumAnalyzerMorph methodsFor: 'stepping and presenter' stamp: 'jm 9/8/1999 19:05'!
step
	"Update the record light, level meter, and display."

	| w |
	"update the record light and level meter"
	soundInput isRecording
		ifTrue: [statusLight color: Color yellow]
		ifFalse: [statusLight color: Color gray].
	w := ((121 * soundInput meterLevel) // 100) max: 1.
	levelMeter width ~= w ifTrue: [levelMeter width: w].

	"update the display if any data is available"
	self updateDisplay.
! !

!SpectrumAnalyzerMorph methodsFor: 'testing' stamp: 'jm 9/7/1999 22:26'!
stepTime

	^ 0
! !

!SpectrumAnalyzerMorph methodsFor: 'stepping and presenter' stamp: 'jm 9/8/1999 15:12'!
stop
	"Stop displaying sound data."

	soundInput stopRecording.
! !

!SpectrumAnalyzerMorph methodsFor: 'stepping and presenter' stamp: 'jm 9/6/1999 12:12'!
stopStepping
	"Turn off recording."

	super stopStepping.
	soundInput stopRecording.
! !

!SpectrumAnalyzerMorph methodsFor: 'private' stamp: 'jm 9/8/1999 19:39'!
updateDisplay
	"Update the display if any data is available."

	| buf bufCount |
	soundInput bufferCount = 0 ifTrue: [^ self].

	graphMorph ifNotNil: [
		[soundInput bufferCount > 0] whileTrue: [
			"skip to the most recent buffer"
			buf := soundInput nextBufferOrNil].
		^ self processBuffer: buf].

	sonogramMorph ifNotNil: [
		"at small buffer sizes we have to update the sonogram in
		 batches or we may get behind; shoot for 8 updates/second"
		bufCount := (soundInput samplingRate / (8 * soundInput bufferSize)) truncated max: 1.
		[bufCount > 0 and: [soundInput bufferCount > 0]] whileTrue: [
			self processBuffer: (soundInput nextBufferOrNil)]].
! !

AlignmentMorph subclass: #WaveEditor
	instanceVariableNames: 'graph samplingRate perceivedFrequency loopEnd loopLength loopCycles possibleLoopStarts keyboard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!WaveEditor commentStamp: '<historical>' prior: 0!
This tool was created to aid in the preparation of LoopedSampledSound objects. It includes support for finding good loop points with a little help from the user.  Namely, the user must identify a good ending point for the loop (typically just before the decay phase begins) and identify one cycle of the waveform. After that, the  "choose loop point" menu command can be invoked to search backwards to find and rank all possible loop starting points. Some experimentation is usually required to find a loop that "breaths" in a natural way.

This tool can also be used as a general viewer of numerical sequences of any kind, such as time-varying functions, FFT data, etc.!

!WaveEditor class methodsFor: 'parts bin' stamp: 'sw 8/2/2001 14:52'!
descriptionForPartsBin
	^ self partName:	'WaveEditor'
		categories:		#('Multimedia')
		documentation:	'A workbench for seing and editing wave forms'! !

!WaveEditor class methodsFor: 'instance creation' stamp: 'jm 9/6/1999 11:21'!
openOn: dataCollection
	"Open a new WaveEditor on the given sequencable collection of data."

	^ (self new data: dataCollection) openInWorld
! !

!WaveEditor methodsFor: 'initialization' stamp: 'dgd 9/19/2003 12:23'!
addControls

	| slider bb r m |
	r := AlignmentMorph newRow.
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r color: bb color; borderWidth: 0; layoutInset: 0.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	r wrapCentering: #topLeft.
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'X';					actionSelector: #delete).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: '<>'; actWhen: #buttonDown;
															actionSelector: #invokeMenu).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Play' translated;				actionSelector: #play).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Play Before' translated;		actionSelector: #playBeforeCursor).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Play After' translated;			actionSelector: #playAfterCursor).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Play Loop' translated;			actionSelector: #playLoop).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Test' translated;				actionSelector: #playTestNote).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Save' translated;				actionSelector: #saveInstrument).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Set Loop End' translated;		actionSelector: #setLoopEnd).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'One Cycle' translated;			actionSelector: #setOneCycle).
	bb := SimpleButtonMorph new target: self; borderColor: Color black.
	r addMorphBack: (bb label: 'Set Loop Start' translated;		actionSelector: #setLoopStart).
	self addMorphBack: r.

	r := AlignmentMorph newRow.
	r color: self color; borderWidth: 0; layoutInset: 0.
	r hResizing: #spaceFill; vResizing: #rigid; extent: 5@20; wrapCentering: #center; cellPositioning: #leftCenter.

	m := StringMorph new contents: 'Index: ' translated.
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: graph; getSelector: #cursor; putSelector: #cursor:;
		growable: false; width: 71; step.
	r addMorphBack: m.

	m := StringMorph new contents: 'Value: ' translated.
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: graph; getSelector: #valueAtCursor; putSelector: #valueAtCursor:;
		growable: false; width: 50; step.
	r addMorphBack: m.

	slider := SimpleSliderMorph new
		color: color;
		extent: 200@2;
		target: self;
		actionSelector: #scrollTime:.
	r addMorphBack: slider.

	m := Morph new color: r color; extent: 10@5.  "spacer"
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: graph; getSelector: #startIndex; putSelector: #startIndex:;
		width: 40; step.
	r addMorphBack: m.

	self addMorphBack: r.

! !

!WaveEditor methodsFor: 'initialization' stamp: 'dgd 9/19/2003 12:23'!
addLoopPointControls

	| r m |
	r := AlignmentMorph newRow.
	r color: self color; borderWidth: 0; layoutInset: 0.
	r hResizing: #spaceFill; vResizing: #rigid; extent: 5@20; wrapCentering: #center; cellPositioning: #leftCenter.

	m := StringMorph new contents: 'Loop end: ' translated.
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: self; getSelector: #loopEnd; putSelector: #loopEnd:;
		growable: false; width: 50; step.
	r addMorphBack: m.

	m := StringMorph new contents: 'Loop length: ' translated.
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: self; getSelector: #loopLength; putSelector: #loopLength:;
		floatPrecision: 0.001;
		growable: false; width: 50; step.
	r addMorphBack: m.

	m := StringMorph new contents: 'Loop cycles: ' translated.
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: self; getSelector: #loopCycles; putSelector: #loopCycles:;
		floatPrecision: 0.001;
		growable: false; width: 50; step.
	r addMorphBack: m.

	m := StringMorph new contents: 'Frequency: ' translated.
	r addMorphBack: m.
	m := UpdatingStringMorph new
		target: self; getSelector: #perceivedFrequency; putSelector: #perceivedFrequency:;
		floatPrecision: 0.001;
		growable: false; width: 50; step.
	r addMorphBack: m.

	self addMorphBack: r.
! !

!WaveEditor methodsFor: 'other' stamp: 'jm 7/30/1998 18:51'!
autoCorrolationBetween: index1 and: index2 length: length
	"Answer the cummulative error between the portions of my waveform starting at the given two indices and extending for the given length. The larger this error, the greater the difference between the two waveforms."

	| data error i1 e |
	data := graph data.
	error := 0.
	i1 := index1.
	index2 to: (index2 + length - 1) do: [:i2 |
		e := (data at: i1) - (data at: i2).
		e < 0 ifTrue: [e := 0 - e].
		error := error + e.
		i1 := i1 + 1].
	^ error
! !

!WaveEditor methodsFor: 'menu' stamp: 'alain.plantec 2/8/2009 22:57'!
chooseLoopStart 

	| bestLoops menu secs choice start |
	possibleLoopStarts ifNil: [
		UIManager default
			informUser: 'Finding possible loop points...' translated
			during: [possibleLoopStarts := self findPossibleLoopStartsFrom: graph cursor]].
	bestLoops := possibleLoopStarts copyFrom: 1 to: (100 min: possibleLoopStarts size).
	menu := CustomMenu new.
	bestLoops do: [:entry |
		secs := ((loopEnd - entry first) asFloat / self samplingRate) roundTo: 0.01.
		menu add: ('{1} ', 'cycles' translated, '; {2} secs' translated format:{entry third. secs}) action: entry].
	choice := menu startUp.
	choice ifNil: [^ self].
	loopCycles := choice third.
	start := self fractionalLoopStartAt: choice first.
	self loopLength: (loopEnd asFloat - start) + 1.0.
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/5/1998 12:43'!
data: newData

	graph data: newData.
! !

!WaveEditor methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:02'!
defaultBorderWidth
"answer the default border width for the receiver"
	^ 2! !

!WaveEditor methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:02'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color lightGray! !

!WaveEditor methodsFor: 'other' stamp: 'jm 7/28/1998 02:29'!
errorBetween: sampleArray1 and: sampleArray2
	"Answer the cummulative error between the two sample arrays, which are assumed to be the same size."

	| error e |
	error := 0.
	1 to: sampleArray1 size do: [:i |
		e := (sampleArray1 at: i) - (sampleArray2 at: i).
		e < 0 ifTrue: [e := 0 - e].
		error := error + e].
	^ error
! !

!WaveEditor methodsFor: 'other' stamp: 'dgd 2/22/2003 18:59'!
findPossibleLoopStartsFrom: index 
	"Assume loopEnd is one sample before a zero-crossing."

	| r postLoopCycleStart i postLoopCycleLength cycleLength cycleCount err oldI |
	r := OrderedCollection new.

	"Record the start and length of the first cycle after the loop endpoint."
	postLoopCycleStart := loopEnd + 1.	"Assumed to be a zero-crossing."
	i := self zeroCrossingAfter: postLoopCycleStart 
						+ (0.9 * samplingRate / perceivedFrequency) asInteger.
	postLoopCycleLength := i - loopEnd - 1.

	"Step backwards one cycle at a time, using zero-crossings to find the
	 beginning of each cycle, and record the auto-corrolation error between
	 each cycle and the cycle following the loop endpoint. Assume pitch may shift gradually."
	i := self zeroCrossingAfter: postLoopCycleStart 
						- (1.1 * postLoopCycleLength) asInteger.
	cycleLength := postLoopCycleStart - i.
	cycleCount := 1.
	[cycleLength > 0] whileTrue: 
			[err := self 
						autoCorrolationBetween: i
						and: postLoopCycleStart
						length: postLoopCycleLength.
			r add: (Array 
						with: i
						with: err
						with: cycleCount
						with: ((loopEnd - i) asFloat / self samplingRate roundTo: 0.01)).
			oldI := i.
			i := self zeroCrossingAfter: oldI - (1.1 * cycleLength) asInteger.
			cycleLength := oldI - i.	"will be zero when start of data is encountered"
			cycleCount := cycleCount + 1].
	r := r asSortedCollection: [:e1 :e2 | (e1 second) < (e2 second)].
	^r asArray! !

!WaveEditor methodsFor: 'other' stamp: 'jm 8/17/1998 11:27'!
fractionalLoopStartAt: index
	"Answer the fractional index starting point near the given integral index that results in the closest match with the cycle following the loop end."
	"Note: could do this more efficiently by sliding downhill on the error curve to find lowest error."

	| oneCycle w1 minErr w2 err bestIndex |
	oneCycle := (samplingRate / perceivedFrequency) rounded.
	w1 := self interpolatedWindowAt: loopEnd + 1 width: oneCycle.
	minErr := SmallInteger maxVal.
	((index - 2) max: 1) to: ((index + 2) min: graph data size) by: 0.01 do: [:i |
		w2 := self interpolatedWindowAt: i width: oneCycle.
		err := self errorBetween: w1 and: w2.
		err < minErr ifTrue: [
			bestIndex := i.
			minErr := err]].
	^ bestIndex
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/3/2004 18:16'!
graph

	^ graph
! !

!WaveEditor methodsFor: 'initialization' stamp: 'nk 6/22/2004 16:22'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	samplingRate := SoundPlayer samplingRate.
	loopEnd := loopLength := 0.
	loopCycles := 1.
	perceivedFrequency := 0.
	"zero means unknown"
	self extent: 5 @ 5;
		 listDirection: #topToBottom;
		 wrapCentering: #topLeft;
		 hResizing: #shrinkWrap;
		 vResizing: #shrinkWrap;
		 layoutInset: 3.
	graph := GraphMorph new extent: 450 @ 100.

	graph cursor: 0.
	graph cursorColorAtZeroCrossings: Color blue.
	self addControls.
	self addLoopPointControls.
	self addMorphBack: graph.
	self
		addMorphBack: (Morph
				newBounds: (0 @ 0 extent: 0 @ 3)
				color: Color transparent).
	self addMorphBack: (keyboard := PianoKeyboardMorph new).
	self sound: (SampledSound soundNamed: 'croak').
! !

!WaveEditor methodsFor: 'other' stamp: 'jm 7/31/1998 11:54'!
interpolatedWindowAt: index width: nSamples
	"Return an array of N samples starting at the given index in my data."

	| scale data baseIndex scaledFrac scaledOneMinusFrac prevSample nextSample v |
	scale := 10000.
	data := graph data.
	index isInteger
		ifTrue: [^ (index to: index + nSamples - 1) collect: [:i | data at: i]].
	baseIndex := index truncated.
	scaledFrac := ((index asFloat - baseIndex) * scale) truncated.
	scaledOneMinusFrac := scale - scaledFrac.
	prevSample := data at: baseIndex.
	^ (baseIndex + 1 to: baseIndex + nSamples) collect: [:i |
		nextSample := data at: i.
		v := ((nextSample * scaledFrac) + (prevSample * scaledOneMinusFrac)) // scale.
		prevSample := nextSample.
		v].
! !

!WaveEditor methodsFor: 'menu' stamp: 'dgd 9/19/2003 12:47'!
invokeMenu
	"Invoke a menu of additonal functions for this WaveEditor."

	| aMenu |
	aMenu := CustomMenu new.
	aMenu addList:	{
		{'play straight through' translated.	#play}.
		{'play before cursor' translated.	#playBeforeCursor}.
		{'play after cursor' translated.		#playAfterCursor}.
		{'play test note' translated.			#playTestNote}.
		{'play loop' translated.				#playLoop}.
		{'trim before cursor' translated.	#trimBeforeCursor}.
		{'trim after cursor' translated.		#trimAfterCursor}.
		{'choose loop start' translated.		#chooseLoopStart}.
		{'jump to loop start' translated.		#jumpToLoopStart}.
		{'jump to loop end' translated.		#jumpToLoopEnd}.
		{'make unlooped' translated.		#setUnlooped}.
		{'make unpitched' translated.		#setUnpitched}.
		{'show envelope' translated.		#showEnvelope}.
		{'show FFT' translated.				#showFFTAtCursor}}.
	aMenu invokeOn: self defaultSelection: nil.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/3/1998 12:42'!
jumpToLoopEnd

	graph cursor: loopEnd; centerCursor.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/17/1998 10:09'!
jumpToLoopStart

	graph cursor: (loopEnd - loopLength) truncated; centerCursor.
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/7/1998 09:48'!
loopCycles

	^ loopCycles
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/8/1998 20:21'!
loopCycles: aNumber

	loopCycles := aNumber.
	self loopLength: loopLength.  "updates frequency"
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 6/30/1998 17:20'!
loopEnd

	^ loopEnd
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/31/1998 14:01'!
loopEnd: aNumber

	loopEnd := (aNumber asInteger max: 1) min: graph data size.
	possibleLoopStarts := nil.
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/7/1998 08:38'!
loopLength

	^ loopLength
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/8/1998 21:03'!
loopLength: aNumber

	loopLength := aNumber.
	((loopCycles > 0) and: [loopLength > 0]) ifTrue: [
		perceivedFrequency := samplingRate asFloat * loopCycles / loopLength].

! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 8/17/1998 10:08'!
loopStart

	^ (loopEnd - loopLength) truncated + 1
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 8/17/1998 20:46'!
loopStart: index

	| start len |
	start := self fractionalLoopStartAt: index.
	len := (loopEnd asFloat - start) + 1.0.
	loopCycles := (len / (samplingRate / perceivedFrequency)) rounded.
	self loopLength: len.
! !

!WaveEditor methodsFor: 'menu' stamp: 'nk 6/22/2004 18:34'!
makeLoopedSampledSound
	| data end snd basePitch |
	data := graph data.
	snd := (loopEnd = 0 or: [loopLength = 0]) 
				ifTrue: 
					["save as unlooped"

					basePitch := perceivedFrequency = 0 
						ifTrue: [100.0]
						ifFalse: [perceivedFrequency].
					LoopedSampledSound new 
						unloopedSamples: data
						pitch: basePitch
						samplingRate: samplingRate]
				ifFalse: 
					[end := (loopEnd min: data size) max: 1.
					basePitch := samplingRate * loopCycles / loopLength.
					LoopedSampledSound new 
						samples: data
						loopEnd: end
						loopLength: end
						pitch: basePitch
						samplingRate: samplingRate].
	snd addReleaseEnvelope.
	^snd! !

!WaveEditor methodsFor: 'other' stamp: 'gm 2/27/2003 23:31'!
normalize: sampleArray 
	"Return a copy of the given sample array scaled to use the maximum 16-bit sample range. Remove any D.C. offset."

	| max scale out |
	max := 0.
	sampleArray do: 
			[:s | 
			max := max max: s abs].
	scale := ((1 << 15) - 1) asFloat / max.
	out := sampleArray species new: sampleArray size.
	1 to: sampleArray size
		do: [:i | out at: i put: (scale * (sampleArray at: i)) truncated].
	^out! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 8/17/1998 20:31'!
perceivedFrequency

	^ perceivedFrequency

! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 8/17/1998 20:32'!
perceivedFrequency: aNumber

	perceivedFrequency := aNumber.
	(loopCycles > 0) ifTrue: [
		loopLength := samplingRate asFloat * loopCycles / perceivedFrequency].
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 7/9/1998 09:23'!
play

	graph data size < 2 ifTrue: [^ self].
	(SampledSound samples: graph data samplingRate: samplingRate) play.

! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 6/30/1998 17:40'!
playAfterCursor

	self playFrom: graph cursor to: graph data size.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 6/30/1998 17:39'!
playBeforeCursor

	self playFrom: 1 to: graph cursor.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 6/30/1998 17:02'!
playFrom: start to: end

	| sz i1 i2 snd |
	sz := graph data size.
	i1 := ((start + 1) min: sz) max: 1.
	i2 := ((end + 1) min: sz) max: i1.
	(i1 + 2) >= i2 ifTrue: [^ self].
	snd := SampledSound
		samples: (graph data copyFrom: i1 to: i2)
		samplingRate: samplingRate.
	snd play.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/17/1998 11:39'!
playLoop

	| sz i1 i2 snd len |
	sz := graph data size.
	i1 := ((loopEnd - loopLength) truncated min: sz) max: 1.
	i2 := (loopEnd min: sz) max: i1.
	len := (i2 - i1) + 1.
	len < 2 ifTrue: [^ self].

	snd := LoopedSampledSound new
		samples: (graph data copyFrom: i1 to: i2)
		loopEnd: len
		loopLength: loopLength
		pitch: 100.0
		samplingRate: samplingRate.

	"sustain for the longer of four loops or two seconds"
	snd setPitch: 100.0
		dur: (((4.0 * loopLength) / samplingRate) max: 2.0)
		loudness: 0.5.
	snd play.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/17/1998 11:38'!
playTestNote

	| data end snd loopDur dur |
	(loopEnd = 0 or: [loopLength = 0]) ifTrue: [^ self].
	data := graph data.
	end := (loopEnd min: data size) max: 1.
	snd := LoopedSampledSound new
		samples: data loopEnd: end loopLength: loopLength
		pitch: 100.0 samplingRate: samplingRate.

	loopDur := (4.0 * loopLength / samplingRate) max: 2.0.  "longer of 4 loops or 2 seconds"
	dur := (data size / samplingRate) + loopDur.
	(snd
		addReleaseEnvelope;
		setPitch: 100.0 dur: dur loudness: 0.5) play.
! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/4/1998 10:44'!
samplingRate

	^ samplingRate

! !

!WaveEditor methodsFor: 'accessing' stamp: 'jm 7/4/1998 10:44'!
samplingRate: samplesPerSecond

	samplingRate := samplesPerSecond.
! !

!WaveEditor methodsFor: 'menu' stamp: 'rbb 3/1/2005 11:20'!
saveInstrument

	| name |
	name := UIManager default request: 'Instrument name?' translated.
	name isEmpty ifTrue: [^ self].
	AbstractSound soundNamed: name put: self makeLoopedSampledSound.
! !

!WaveEditor methodsFor: 'other' stamp: 'jm 6/30/1998 16:11'!
scrollTime: relativeValue

	graph startIndex: relativeValue * (graph data size - (graph width // 2)).
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 7/31/1998 11:06'!
setLoopEnd

	graph cursor: (self zeroCrossingAfter: graph cursor) - 1.
	self loopEnd: graph cursor.

! !

!WaveEditor methodsFor: 'menu' stamp: 'dgd 9/19/2003 12:48'!
setLoopStart
	"Assuming that the loop end and approximate frequency have been set, this method uses the current cursor position to determine the loop length and the number of cycles."

	| start len |
	start := graph cursor.
	((start >= loopEnd) or: [perceivedFrequency = 0]) ifTrue: [
		^ self inform:
'Please set the loop end and the approximate frequency
first, then position the cursor one or more cycles
before the loop end and try this again.' translated].
	len := (loopEnd - start) + 1.
	loopCycles := (len / (samplingRate / perceivedFrequency)) rounded.
	self loopLength: len.

! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/1/1998 11:05'!
setOneCycle
	"Set the approximate frequency based on a single cycle specified by the user. To use this, first set the loop end, then place the cursor one full cycle before the loop end and invoke this method."

	| len |
	len := loopEnd - graph cursor.
	len > 0 ifTrue: [
		loopCycles := 1.
		self loopLength: len].
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/1/1998 11:05'!
setUnlooped
	"Make this sound play straight through without looping."

	loopLength := 0.
	loopCycles := 1.
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 8/1/1998 11:05'!
setUnpitched
	"Make this instrument be unpitched and unlooped. Suitable for percussive sounds that should not be pitch-shifted when played. By convention, such sounds are played at a pitch of 100.0 to obtain their original pitch."

	loopLength := 0.
	loopCycles := 0.
	perceivedFrequency := 100.0.
! !

!WaveEditor methodsFor: 'menu' stamp: 'di 7/16/1999 08:30'!
showEnvelope
	"Show an envelope wave constructed by collecting the maximum absolute value of the samples in fixed-size time windows of mSecsPerQuantum."

	| data mSecsPerQuantum samplesPerQuantum result endOfQuantum maxThisQuantum s nSamples |
	data := graph data.
	mSecsPerQuantum := 10.
	samplesPerQuantum := (mSecsPerQuantum / 1000.0) * self samplingRate.
	result := WriteStream on: (Array new: data size // samplesPerQuantum).
	endOfQuantum := samplesPerQuantum.
	maxThisQuantum := 0.
	nSamples := (data isKindOf: SoundBuffer)
		ifTrue: [data monoSampleCount]
		ifFalse: [data size].
	1 to: nSamples do: [:i |
		i asFloat > endOfQuantum ifTrue: [
			result nextPut: maxThisQuantum.
			maxThisQuantum := 0.
			endOfQuantum := endOfQuantum + samplesPerQuantum].
		s := data at: i.
		s < 0 ifTrue: [s := 0 - s].
		s > maxThisQuantum ifTrue: [maxThisQuantum := s]].
	WaveEditor openOn: result contents.
! !

!WaveEditor methodsFor: 'menu' stamp: 'dgd 9/19/2003 12:48'!
showFFTAtCursor

	| data start availableSamples nu n fft r |
	data := graph data.
	start := graph cursor max: 1.
	availableSamples := (data size - start) + 1.
	nu := 12.
	nu > (availableSamples highBit - 1) ifTrue:
		[^ self inform: 'Too few samples after the cursor to take an FFT.' translated].
	n := 2 raisedTo: nu.
	fft := FFT new nu: nu.
	fft realData: ((start to: start + n - 1) collect: [:i | data at: i]).
	fft transformForward: true.
	r := (1 to: n // 2) collect:
		[:i | ((fft realData at: i) squared + (fft imagData at: i) squared) sqrt].
	WaveEditor openOn: r.

! !

!WaveEditor methodsFor: 'accessing' stamp: 'nk 6/22/2004 16:17'!
sound: aSound
	| buffer |
	buffer := aSound samples mergeStereo.
	graph data: buffer.
	loopLength := loopEnd := buffer size.
	self samplingRate: aSound originalSamplingRate.
	loopCycles :=  buffer size / aSound originalSamplingRate * 400.
	perceivedFrequency := 400.
! !

!WaveEditor methodsFor: 'stepping and presenter' stamp: 'di 6/21/1999 17:59'!
step

	keyboard soundPrototype: self makeLoopedSampledSound! !

!WaveEditor methodsFor: 'other' stamp: 'jm 9/20/1998 09:13'!
stretch: sampleArray by: stretchFactor
	"Return an array consisting of the given samples \stretched in time by the given factor."

	| out end incr i frac index |
	out := OrderedCollection new: (stretchFactor * sampleArray size) asInteger + 1.
	end := (sampleArray size - 1) asFloat.
	incr := 1.0 / stretchFactor.
	i := 1.0.
	[i < end] whileTrue: [
		frac := i fractionPart.
		index := i truncated.
		i := i + incr.
		out addLast:
			(((1.0 - frac) * (sampleArray at: index)) + (frac * (sampleArray at: index + 1))) rounded].
	^ out asArray
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 6/30/1998 17:48'!
trimAfterCursor

	graph data: (graph data copyFrom: 1 to: graph cursor).
! !

!WaveEditor methodsFor: 'menu' stamp: 'jm 6/30/1998 17:52'!
trimBeforeCursor

	graph data: (graph data copyFrom: graph cursor to: graph data size).
	graph cursor: 1.

! !

!WaveEditor methodsFor: 'other' stamp: 'jm 8/10/1998 15:08'!
zeroCrossingAfter: index
	"Find the index of the next negative-to-non-negative transition at or after the current index. The result is the index, i, of a zero crossing such that the sample at i-1 is negative and the sample at i is zero or positive. Answer the index of the last sample if the end of the array is encountered before finding a zero crossing."

	| data end i |
	data := graph data.
	end := data size.
	index <= 1 ifTrue: [^ 1].
	i := index - 1.
	[(i <= end) and: [(data at: i) >= 0]] whileTrue: [i := i + 1].  "find next negative sample"
	i >= end ifTrue: [^ end].

	i := i + 1.
	[(i <= end) and: [(data at: i) < 0]] whileTrue: [i := i + 1].  "find next non-negative sample"
	^ i
! !

RectangleMorph subclass: #AbstractMediaEventMorph
	instanceVariableNames: 'startTimeInScore endTimeInScore'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!AbstractMediaEventMorph commentStamp: '<historical>' prior: 0!
An abstract representation of media events to be placed in a PianoRollScoreMorph (or others as they are developed)!

!AbstractMediaEventMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:36'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!AbstractMediaEventMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:24'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color paleYellow! !

!AbstractMediaEventMorph methodsFor: 'as yet unclassified' stamp: 'RAA 12/7/2000 12:58'!
endTime

	^endTimeInScore ifNil: [startTimeInScore + 100]! !

!AbstractMediaEventMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:38'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self layoutPolicy: TableLayout new;
	  listDirection: #leftToRight;
	  wrapCentering: #topLeft;
	  hResizing: #shrinkWrap;
	  vResizing: #shrinkWrap;
	  layoutInset: 2;
	  rubberBandCells: true! !

AbstractMediaEventMorph subclass: #SoundEventMorph
	instanceVariableNames: 'sound'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SoundEventMorph methodsFor: 'piano rolls' stamp: 'RAA 12/7/2000 12:36'!
addMorphsTo: morphList pianoRoll: pianoRoll eventTime: t betweenTime: leftTime and: rightTime

	| startX lengthInTicks endX |

	startTimeInScore > rightTime ifTrue: [^ self].  
	lengthInTicks := pianoRoll scorePlayer ticksForMSecs: sound duration * 1000.0.
	startTimeInScore + lengthInTicks < leftTime ifTrue: [^ self].
	startX := pianoRoll xForTime: startTimeInScore.
	endX := pianoRoll xForTime: startTimeInScore + lengthInTicks.
	morphList add: 
		(self left: startX; width: endX - startX).

! !

!SoundEventMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color lightGreen! !

!SoundEventMorph methodsFor: 'piano rolls' stamp: 'RAA 12/7/2000 12:29'!
encounteredAtTime: ticks inScorePlayer: scorePlayer atIndex: index inEventTrack: track secsPerTick: secsPerTick

	"hack... since we are called from within the SoundPlayer loop, the Semaphore will
	block attempts to play directly from here"
	WorldState addDeferredUIMessage: [sound play].! !

!SoundEventMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:39'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	self height: 10! !

!SoundEventMorph methodsFor: 'piano rolls' stamp: 'RAA 12/9/2000 18:48'!
justDroppedIntoPianoRoll: newOwner event: evt
	
	| startX lengthInTicks endX |

	super justDroppedIntoPianoRoll: newOwner event: evt.

	startTimeInScore := newOwner timeForX: self left.
	lengthInTicks := newOwner scorePlayer ticksForMSecs: sound duration * 1000.0.
	endTimeInScore := startTimeInScore + lengthInTicks.

	endTimeInScore > newOwner scorePlayer durationInTicks ifTrue:
		[newOwner scorePlayer updateDuration].

	startX := newOwner xForTime: startTimeInScore.
	endX := newOwner xForTime: endTimeInScore.
	self width: endX - startX.

! !

!SoundEventMorph methodsFor: 'caching' stamp: 'RAA 12/8/2000 09:52'!
releaseCachedState

	super releaseCachedState.
	sound := sound compressWith: GSMCodec.
! !

!SoundEventMorph methodsFor: 'as yet unclassified' stamp: 'RAA 12/7/2000 13:01'!
sound: aSound

	sound := aSound.
	self setBalloonText: 'a sound of duration ',(sound duration roundTo: 0.1) printString,' seconds'.! !

RectangleMorph subclass: #CodecDemoMorph
	instanceVariableNames: 'codecClassName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!CodecDemoMorph methodsFor: 'layout' stamp: 'jm 2/4/1999 12:37'!
acceptDroppingMorph: aMorph event: evt

	| codecClass |
	'None' = codecClassName
		ifTrue: [aMorph sound play]
		ifFalse: [
			codecClass := Smalltalk at: codecClassName ifAbsent: [^ self].
			(codecClass new compressAndDecompress: aMorph sound) play].
	aMorph position: self topRight + (10@0).
! !

!CodecDemoMorph methodsFor: 'menus' stamp: 'dgd 8/30/2003 21:17'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'select codec' translated action: #selectCodec.
! !

!CodecDemoMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/4/1999 12:36'!
codecClassName: aStringOrSymbol

	| label |
	codecClassName := aStringOrSymbol asSymbol.
	self removeAllMorphs.
	label := StringMorph contents: aStringOrSymbol.
	label position: self position + (5@5).
	self addMorph: label.
	label lock: true.
	self extent: label extent + (10@10).
! !

!CodecDemoMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:20'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color
		r: 1.0
		g: 0.806
		b: 0.677! !

!CodecDemoMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:20'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	self codecClassName: 'MuLawCodec'! !

!CodecDemoMorph methodsFor: 'as yet unclassified' stamp: 'alain.plantec 2/8/2009 21:39'!
selectCodec

	| codecs newCodec |
	codecs := (SoundCodec allSubclasses collect: [:c | c name]) asSortedCollection.
	codecs add: 'None'.
	newCodec := UIManager default 
		chooseFrom: codecs values: codecs title: 'Codec:' translated.
	newCodec ifNil: [^ self].
	self codecClassName: newCodec.
! !

!CodecDemoMorph methodsFor: 'dropping/grabbing' stamp: 'marcus.denker 11/18/2008 14:51'!
wantsDroppedMorph: aMorph event: evt

	^ false
! !

RectangleMorph subclass: #EnvelopeEditorMorph
	instanceVariableNames: 'sound soundName envelope hScale vScale graphArea pixPerTick limits limitXs limitHandles line prevMouseDown sampleDuration showAllEnvelopes denominator keyboard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!EnvelopeEditorMorph class methodsFor: 'as yet unclassified' stamp: 'di 5/15/1998 09:49'!
openOn: aSound title: aString
	"EnvelopeEditorMorph openOn: (AbstractSound soundNamed: 'brass1') copy title: 'brass1'"
	(self basicNew initOnSound: aSound title: aString) openInWorld! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'dgd 2/22/2003 14:23'!
acceptGraphPoint: p at: index 
	| ms val points whichLim linePoint other boundedP |
	boundedP := p adhereTo: graphArea bounds.
	ms := self msFromX: boundedP x.
	points := envelope points.
	ms := self 
				constrain: ms
				adjacentTo: index
				in: points.
	(index = 1 or: [(whichLim := limits indexOf: index) > 0]) 
		ifTrue: 
			["Limit points must not move laterally"

			ms := (points at: index) x].
	val := self valueFromY: boundedP y.
	points at: index put: ms @ val.
	linePoint := (self xFromMs: ms) @ (self yFromValue: val).
	(whichLim notNil and: [whichLim between: 1 and: 2]) 
		ifTrue: 
			["Loop start and loop end must be tied together"

			other := limits at: 3 - whichLim.	" 1 <--> 2 "
			points at: other put: (points at: other) x @ val.
			line verticesAt: other put: (line vertices at: other) x @ linePoint y].
	"Make sure envelope feels the change in points array..."
	envelope 
		setPoints: points
		loopStart: limits first
		loopEnd: (limits second).
	^linePoint! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 7/14/2000 13:08'!
addControls
	| chooser |
	chooser := PopUpChoiceMorph new extent: 110@14;
		contentsClipped: 'Editing: ' , envelope name;
		target: self;
		actionSelector: #chooseFrom:envelopeItem:;
		getItemsSelector: #curveChoices.
	chooser arguments: (Array with: chooser).
	self addMorph: chooser.
	chooser align: chooser bounds topLeft with: graphArea bounds bottomLeft + (0@5).

	chooser := PopUpChoiceMorph new extent: 130@14;
		contentsClipped: 'Timbre: ' , soundName;
		target: self;
		actionSelector: #chooseFrom:soundItem:;
		getItemsSelector: #soundChoices.
	chooser arguments: (Array with: chooser).
	self addMorph: chooser.
	chooser align: chooser bounds topRight with: graphArea bounds bottomRight + (-50@5).
! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'ar 3/17/2001 14:24'!
addCurves
	"Add the polyLine corresponding to the currently selected envelope,
	and possibly all the others, too."
	| verts aLine |
	sound envelopes do:
		[:env | 
		(showAllEnvelopes or: [env == envelope]) ifTrue:
			[verts := env points collect:
				[:p | (self xFromMs: p x) @ (self yFromValue: p y)].
			aLine := EnvelopeLineMorph basicNew
						vertices: verts borderWidth: 1
						borderColor: (self colorForEnvelope: env).
			env == envelope
				ifTrue: [aLine borderWidth: 2.  line := aLine]
				ifFalse: [aLine on: #mouseUp send: #clickOn:evt:from:
							to: self withValue: env.
						self addMorph: aLine]]].
	self addMorph: line  "add the active one last (in front)"! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:19'!
addCustomMenuItems: menu hand: aHandMorph
	super addCustomMenuItems: menu hand: aHandMorph.
	menu addLine.
	envelope updateSelector = #ratio: ifTrue:
		[menu add: 'choose denominator...' translated action: #chooseDenominator:].
	menu add: 'adjust scale...' translated action: #adjustScale:.
	SoundPlayer isReverbOn
		ifTrue: [menu add: 'turn reverb off' translated target: SoundPlayer selector: #stopReverb]
		ifFalse: [menu add: 'turn reverb on' translated target: SoundPlayer selector: #startReverb].
	menu addLine.
	menu add: 'get sound from lib' translated action: #chooseSound:.
	menu add: 'put sound in lib' translated action: #saveSound:.
	menu add: 'read sound from disk...' translated action: #readFromDisk:.
	menu add: 'save sound on disk...' translated action: #saveToDisk:.
	menu add: 'save library on disk...' translated action: #saveLibToDisk:.
! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'JMV 1/29/2001 10:57'!
addEnvelopeNamed: envName
	| points env |
	points := OrderedCollection new.
	points add: 0@0.0;
		add: (envelope points at: envelope loopStartIndex) x@1.0;
		add: (envelope points at: envelope loopEndIndex) x@1.0;
		add: (envelope points last) x@0.0.
	envName = 'volume' ifTrue:
		[env := VolumeEnvelope points: points loopStart: 2 loopEnd: 3.
		env target: sound; scale: 0.7].
	envName = 'modulation' ifTrue:
		[env := Envelope points: (points collect: [:p | p x @ 0.5])
						loopStart: 2 loopEnd: 3.
		env target: sound; updateSelector: #modulation:;
			scale: sound modulation*2.0].
	envName = 'pitch' ifTrue:
		[env := PitchEnvelope points: (points collect: [:p | p x @ 0.5])
						loopStart: 2 loopEnd: 3.
		env target: sound; updateSelector: #pitch:; scale: 0.5].
	envName = 'random pitch:' ifTrue:
		[env := RandomEnvelope for: #pitch:.
		points := OrderedCollection new.
		points add: 0@(env delta * 5 + 0.5);
			add: (envelope points at: envelope loopStartIndex) x@(env highLimit - 1 * 5 + 0.5);
			add: (envelope points at: envelope loopEndIndex) x@(env highLimit - 1 * 5 + 0.5);
			add: (envelope points last) x@(env lowLimit - 1 * 5 + 0.5).
		env setPoints: points loopStart: 2 loopEnd: 3.
		env target: sound. ].
	envName = 'ratio' ifTrue:
		[denominator := 9999.  "No gridding"
		env := Envelope points: (points collect: [:p | p x @ 0.5])
						loopStart: 2 loopEnd: 3.
		env target: sound; updateSelector: #ratio:;
			scale: sound ratio*2.0].
	env ifNotNil:
		[sound addEnvelope: env.
		self editEnvelope: env]! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'ar 3/17/2001 14:25'!
addHandlesIn: frame
	| handle |
	handle := PolygonMorph
		vertices: (Array with: 0@0 with: 8@0 with: 4@8)
		color: Color orange borderWidth: 1 borderColor: Color black.
	handle addMorph: ((RectangleMorph
			newBounds: ((self handleOffset: handle)-(2@0) extent: 1@(graphArea height-2))
			color: Color orange) borderWidth: 0).

	limitHandles := Array with: handle with: handle veryDeepCopy with: handle veryDeepCopy.
	1 to: limitHandles size do:
		[:i | handle := limitHandles at: i.
		handle on: #mouseDown
				send: #limitHandleMove:event:from:
				to: self withValue: i.
		handle on: #mouseMove
				send: #limitHandleMove:event:from:
				to: self withValue: i.
		self addMorph: handle.
		handle position: ((self xFromMs: 
			(envelope points at: (limits at: i)) x) @ 
				(graphArea top)) - (self handleOffset: handle)]! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 9/4/1998 15:49'!
addKeyboard
	keyboard := PianoKeyboardMorph new soundPrototype: sound.
	keyboard align: keyboard bounds bottomCenter with: bounds bottomCenter - (0@4).
	self addMorph: keyboard! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:20'!
adjustScale: evt 
	| scaleString oldScale baseValue |
	oldScale := envelope scale.
	scaleString := UIManager default request: 'Enter the new full-scale value...' translated
				initialAnswer: oldScale printString.
	scaleString isEmpty ifTrue: [^self].
	envelope scale: (Number readFrom: scaleString) asFloat.
	baseValue := envelope updateSelector = #pitch: ifTrue: [0.5] ifFalse: [0.0].
	envelope 
		setPoints: (envelope points collect: 
					[:p | 
					p x @ ((p y - baseValue) * oldScale / envelope scale + baseValue min: 1.0
								max: 0.0)])
		loopStart: (limits first)
		loopEnd: (limits second).
	self buildView! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 9/4/1998 15:56'!
buildGraphAreaIn: frame
	| r y |
	graphArea := RectangleMorph
		newBounds: ((frame left + 40) @ (frame top + 40)
		corner: (frame right+1) @ (frame bottom - 60))
		color: Color lightGreen lighter lighter.
	graphArea borderWidth: 1.
	self addMorph: graphArea.
	(envelope updateSelector = #pitch: and: [envelope scale <= 2.0]) ifTrue:
		["Show half-steps"
		r := graphArea innerBounds.
		0.0 to: 1.0 by: 1.0/12.0/envelope scale do:
			[:val |
			y := self yFromValue: val.
			graphArea addMorph: ((RectangleMorph
					newBounds: (r left@y extent: r width@1)
					color: Color veryLightGray)
						borderWidth: 0)]].
	(envelope updateSelector = #ratio: and: [denominator ~= 9999]) ifTrue:
		["Show denominator gridding"
		r := graphArea innerBounds.
		(0.0 to: 1.0 by: 1.0/denominator/envelope scale) do:
			[:v |
			y := self yFromValue: v.
			graphArea addMorph: ((RectangleMorph
					newBounds: (r left@y extent: r width@1)
					color: Color veryLightGray)
						borderWidth: 0)]].
! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'yo 9/15/2008 10:44'!
buildScalesIn: frame
	| env |
	env := envelope.
	pixPerTick := graphArea width // (self maxTime//10) max: 1.
	"hminortick := ( 1 + ( self maxTime // 800 ) ) * 10.
	hmajortick := ( 1 + ( self maxTime // 800 ) ) * 100."
	hScale := (Morph newBounds: ((graphArea left)@(frame top) corner: (self xFromMs: self maxTime)@(graphArea top - 1))) color: Color lightGreen.
"	hScale := (ScaleMorph newBounds: ((graphArea left)@(frame top) corner: (self xFromMs: self maxTime)@(graphArea top - 1)))
		start: 0 stop: self maxTime
		minorTick: hminortick minorTickLength: 3
		majorTick: hmajortick majorTickLength: 10
		caption: 'milliseconds' tickPrintBlock: [:v | v printString].
"
	self addMorph: hScale.
	vScale := (Morph newBounds: (0@0 extent: (graphArea height)@(graphArea left - frame left))) color: Color lightGreen.
	"vScale := ScaleMorph newBounds: (0@0 extent: (graphArea height)@(graphArea left - frame left))."
	env name = 'pitch'
		ifTrue:
		[env scale >= 2.0
			ifTrue:
			[vScale start: 0 stop: env scale
				minorTick: env scale / 24 minorTickLength: 3
				majorTick: env scale / 2.0 majorTickLength: 10
				caption: 'pitch (octaves)'
				tickPrintBlock: [:v | (v-(env scale/2)) asInteger printString]]
			ifFalse:
			[vScale start: 0 stop: env scale
				minorTick: 1.0/48.0 minorTickLength: 3
				majorTick: 1.0/12.0 majorTickLength: 10
				caption: 'pitch (half-steps)'
				tickPrintBlock: [:v | (v-(env scale/2)*12) rounded printString]]]
		ifFalse: [
			env name = 'random pitch:'
				ifTrue: [
					vScale start: 0.9 stop: 1.1
						minorTick: 0.2 / 50.0 minorTickLength: 3
						majorTick: 0.2 / 5.0 majorTickLength: 10
						caption: env name
						tickPrintBlock: [:v | v printString]]
				ifFalse: [
					vScale start: 0 stop: env scale
						minorTick: env scale / 50.0 minorTickLength: 3
						majorTick: env scale / 5.0 majorTickLength: 10
						caption: env name
						tickPrintBlock: [:v | v printString]].
		].
	vScale := TransformationMorph new asFlexOf: vScale.
	vScale angle: Float pi / 2.0.
	self addMorph: vScale.
	vScale position: (frame left)@(graphArea top-1) - (3@1).
! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 9/4/1998 13:16'!
buildView
	| frame |
	self color: Color lightGreen.
	self removeAllMorphs.
	frame := self innerBounds.
	self buildGraphAreaIn: frame.
	self buildScalesIn: frame.
	self addHandlesIn: frame.
	self addCurves.
	line addHandles.
	self addControls.
	self addKeyboard! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'RAA 6/12/2000 08:58'!
chooseDenominator: evt
	| menu |
	menu := MenuMorph new.
	(Integer primesUpTo: 30) do:
		[:i |
		menu add: i printString
			target: self selector: #setDenominator:
			argument: i].
	menu addLine.
	menu add: 'none' target: self selector: #setDenominator: argument: 9999.
	menu popUpEvent: evt in: self world! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 2/3/98 16:50'!
chooseEnvelope: choice
	| name |
	(choice beginsWith: 'edit ') ifTrue:
		[name := choice copyFrom: 'edit ' size+1 to: choice size.
		^ self editEnvelope: (sound envelopes detect:
				[:env | env name = name])].
	(choice beginsWith: 'add ') ifTrue:
		[name := choice copyFrom: 'add ' size+1 to: choice size.
		^ self addEnvelopeNamed: name].
	(choice beginsWith: 'remove ') ifTrue:
		[^ self removeEnvelope  "the current one"].
! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 2/3/98 19:14'!
chooseFrom: chooserMorph envelopeItem: item
	| name |
	(item beginsWith: 'edit ') ifTrue:
		[name := item copyFrom: 'edit ' size+1 to: item size.
		self editEnvelope: (sound envelopes detect:
				[:env | env name = name])].
	(item beginsWith: 'add ') ifTrue:
		[name := item copyFrom: 'add ' size+1 to: item size.
		self addEnvelopeNamed: name].
	(item beginsWith: 'remove ') ifTrue:
		[self removeEnvelope  "the current one"].
	chooserMorph contentsClipped: envelope name! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 7/14/2000 13:03'!
chooseFrom: chooserMorph soundItem: item
	self editSoundNamed: item.
! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 7/14/2000 12:42'!
chooseSound: evt
	| menu |
	menu := MenuMorph new.
	menu add: 'new...' target: self selector: #editNewSound.
	menu addLine.
	AbstractSound soundNames do:
		[:name |
		menu add: name
			target: self selector: #editSoundNamed:
			argument: name].
	menu popUpEvent: evt in: self world! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'ar 3/17/2001 14:24'!
clickOn: env evt: anEvent from: aLine
	self editEnvelope: env! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'ar 3/18/2001 17:27'!
clickOnLine: arg1 evt: arg2 envelope: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self clickOn: arg1 evt: arg2 from: arg3! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 2/3/98 16:50'!
colorForEnvelope: env
	| name index |
	name := env name.
	index := #('volume' 'modulation' 'pitch' 'ratio') indexOf: name
				ifAbsent: [5].
	^ Color perform: (#(red green blue magenta black) at: index)! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'di 1/29/98 13:06'!
constrain: xVal adjacentTo: ix in: points
	"Return xVal, restricted between points adjacent to vertX"
	| newVal |
	newVal := xVal.
	ix > 1 ifTrue: [newVal := newVal max: (points at: ix-1) x].
	ix < points size ifTrue: [newVal := newVal min: (points at: ix+1) x].
	^ newVal! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'JMV 1/29/2001 10:58'!
curveChoices
	| extant others |
	extant := sound envelopes collect: [:env | env name].
	others := #('volume' 'modulation' 'pitch' 'random pitch:' 'ratio')
		reject: [:x | (extant includes: x) | ((x = 'pitch') & (extant includes: 'random pitch:')) | ((x = 'random pitch:') & (extant includes: 'pitch')) ].
	^ (extant collect: [:name | 'edit ' , name])
	, (others collect: [:name | 'add ' , name])
	, (sound envelopes size > 1
		ifTrue: [Array with: 'remove ' , envelope name]
		ifFalse: [Array new])! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'dgd 2/22/2003 14:24'!
deletePoint: ix 
	"If the point is a limit point, return false,
	otherwise, delete the point at ix, and return true."

	(limits includes: ix) ifTrue: [^false].
	1 to: limits size
		do: 
			[:i | 
			"Decrease limit indices beyond the deletion"

			(limits at: i) > ix ifTrue: [limits at: i put: (limits at: i) - 1]].
	envelope 
		setPoints: (envelope points 
				copyReplaceFrom: ix
				to: ix
				with: Array new)
		loopStart: (limits first)
		loopEnd: (limits second).
	^true! !

!EnvelopeEditorMorph methodsFor: 'initialization' stamp: 'di 1/31/98 10:41'!
editEnvelope: env
	envelope := env.
	limits := Array with: envelope loopStartIndex
				with: envelope loopEndIndex
				with: envelope points size.
	limitXs := limits collect: [:i | (envelope points at: i) x].
	self buildView! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 1/30/98 22:58'!
editNewSound
	| known i |
	known := AbstractSound soundNames.
	i := 0.
	[soundName := 'unnamed' , i printString.
	known includes: soundName]
		whileTrue: [i := 1+1].
	soundName := soundName.
	self editSound: FMSound default copy! !

!EnvelopeEditorMorph methodsFor: 'initialization' stamp: 'alain.plantec 2/6/2009 10:59'!
editSound: aSound

	| p |
	(aSound respondsTo: #envelopes)
		ifFalse: [
			self inform: 'You selected a ' translated, aSound class name, '.', String cr,
				'I can''t handle these kinds of sounds.' translated.
			^self ].
	sound := aSound.
	sound envelopes isEmpty ifTrue: [
		"provide a default volume envelope"
		p := OrderedCollection new.
		p add: 0@0.0; add: 10@1.0; add: 100@1.0; add: 120@0.0.
		sound addEnvelope: (VolumeEnvelope points: p loopStart: 2 loopEnd: 3)].

	self editEnvelope: sound envelopes first.
	keyboard soundPrototype: sound.
! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 7/14/2000 12:44'!
editSoundNamed: name

	name = 'new...' ifTrue: [^ self editNewSound].
	soundName := name.
	self editSound: (AbstractSound soundNamed: soundName) copy! !

!EnvelopeEditorMorph methodsFor: 'geometry' stamp: 'di 9/4/1998 16:03'!
extent: newExtent
	super extent: (newExtent max: (self maxTime//10*3+50 max: 355) @ 284).
	self buildView! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 1/26/98 15:37'!
handleOffset: handle
	"This is the offset from position to the bottom vertex"
	^ (handle width//2+1) @ handle height
! !

!EnvelopeEditorMorph methodsFor: 'initialization' stamp: 'di 9/5/1998 10:40'!
initOnSound: aSound title: title
	sound := aSound.
	soundName := title.
	self initialize.
! !

!EnvelopeEditorMorph methodsFor: 'initialization' stamp: 'di 7/14/2000 12:48'!
initialize
	super initialize.
	prevMouseDown := false.
	showAllEnvelopes := true.
	soundName ifNil: [soundName := 'test'].
	self editSound: (sound ifNil: [FMSound brass1 copy]).
	sound duration: 0.25.
	denominator := 7.
	self extent: 10@10.  "ie the minimum"
! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'dgd 2/22/2003 14:24'!
insertPointAfter: ix 
	"If there is not enough roon (in x) then return false.
	Otherwise insert a point between ix and ix+1 and return true."

	| points pt |
	points := envelope points.
	(points at: ix + 1) x - (points at: ix) x < 20 ifTrue: [^false].
	pt := ((points at: ix + 1) + (points at: ix)) // 2.
	1 to: limits size
		do: 
			[:i | 
			"Increase limit indices beyond the insertion"

			(limits at: i) > ix ifTrue: [limits at: i put: (limits at: i) + 1]].
	envelope 
		setPoints: (points 
				copyReplaceFrom: ix + 1
				to: ix
				with: (Array with: pt))
		loopStart: (limits first)
		loopEnd: (limits second).
	^true! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'ar 3/17/2001 14:24'!
limitHandleMove: index event: evt from: handle
	"index is the handle index = 1, 2 or 3"
	| ix p ms x points limIx |
	ix := limits at: index.  "index of corresponding vertex"
	p := evt cursorPoint adhereTo: graphArea bounds.
	ms := self msFromX: p x + (self handleOffset: handle) x.

	"Constrain move to adjacent points on ALL envelopes"
	sound envelopes do:
		[:env | limIx := env perform:
			(#(loopStartIndex loopEndIndex decayEndIndex) at: index).
		ms := self constrain: ms adjacentTo: limIx in: env points].

	"Update the handle, the vertex and the line being edited"
	x := self xFromMs: ms.
	handle position: (x @ graphArea top) - (self handleOffset: handle).
	line verticesAt: ix put: x @ (line vertices at: ix) y.

	sound envelopes do:
		[:env | limIx := env perform:
			(#(loopStartIndex loopEndIndex decayEndIndex) at: index).
		points := env points.
		points at: limIx put: ms @ (points at: limIx) y.
		env setPoints: points loopStart: env loopStartIndex loopEnd: env loopEndIndex].! !

!EnvelopeEditorMorph methodsFor: 'editing' stamp: 'ar 3/18/2001 17:27'!
limitHandleMoveEvent: arg1 from: arg2 index: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self limitHandleMove: arg1 event: arg2 from: arg3! !

!EnvelopeEditorMorph methodsFor: 'scaling' stamp: 'di 1/27/98 14:40'!
maxTime
	^ (envelope points at: limits last) x + 100! !

!EnvelopeEditorMorph methodsFor: 'scaling' stamp: 'di 1/26/98 14:10'!
msFromX: x
	^ (x - graphArea left)//pixPerTick*10! !

!EnvelopeEditorMorph methodsFor: 'playing' stamp: 'di 2/3/98 17:07'!
playNothing
! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 1/31/98 16:49'!
readFileNamed: fileName
	| snd |
	snd := Compiler evaluate:
		(FileStream readOnlyFileNamed: fileName) contentsOfEntireFile.
	soundName := fileName copyFrom: 1 to: fileName size-4. "---.fmp"
	self editSound: snd! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'RAA 6/12/2000 08:58'!
readFromDisk: evt
	| menu |
	menu := MenuMorph new.
	(FileDirectory default fileNamesMatching: '*.fmp') do:
		[:fileName |
		menu add: fileName
			target: self selector: #readFileNamed:
			argument: fileName].
	menu popUpEvent: evt in: self world! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 11:00'!
removeEnvelope
	(self confirm: 'Really remove ' translated, envelope name , '?')
		ifFalse: [^ self].
	sound removeEnvelope: envelope.
	self editEnvelope: sound envelopes first.! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:20'!
saveLibToDisk: evt
	"Save the library to disk"

	| newName f snd |
	newName := UIManager default request: 'Please confirm name for library...' translated
						initialAnswer: 'MySounds'.
	newName isEmpty ifTrue: [^ self].
	f := FileStream newFileNamed: newName , '.fml'.
	AbstractSound soundNames do:
		[:name | snd := AbstractSound soundNamed: name.
		"snd isStorable" true ifTrue: [f nextChunkPut: 'AbstractSound soundNamed: ' , name , ' put: ' , snd storeString; cr; cr]
			ifFalse: [self inform: name , ' is not currently storable']].
	f close! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:20'!
saveSound: evt
	| newName |
	newName := UIManager default request: 'Please confirm name for save...' translated
						initialAnswer: soundName.
	newName isEmpty ifTrue: [^ self].
	AbstractSound soundNamed: newName put: sound.
	soundName := newName.! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'alain.plantec 2/6/2009 15:20'!
saveToDisk: evt
	| newName f |
	newName := UIManager default request: 'Please confirm name for save...' translated
						initialAnswer: soundName.
	newName isEmpty ifTrue: [^ self].
	f := FileStream newFileNamed: newName , '.fmp'.
	sound storeOn: f.
	f close! !

!EnvelopeEditorMorph methodsFor: 'menu' stamp: 'di 1/31/98 01:36'!
setDenominator: denom
	denominator := denom.
	self buildView! !

!EnvelopeEditorMorph methodsFor: 'initialization' stamp: 'di 11/7/2000 12:45'!
soundBeingEdited

	^ sound! !

!EnvelopeEditorMorph methodsFor: 'construction' stamp: 'di 7/14/2000 12:56'!
soundChoices
	^ #('new...') , AbstractSound soundNames! !

!EnvelopeEditorMorph methodsFor: 'stepping and presenter' stamp: 'di 6/7/1999 15:37'!
step
	| mouseDown hand |
	hand := self world firstHand.
	(bounds containsPoint: hand position) ifFalse: [^ self].

	mouseDown := hand lastEvent redButtonPressed.
	mouseDown not & prevMouseDown ifTrue:
		["Mouse just went up"
		limitXs = (limits collect: [:i | (envelope points at: i) x]) ifFalse:
			["Redisplay after changing limits"
			self editEnvelope: envelope]].
	prevMouseDown := mouseDown! !

!EnvelopeEditorMorph methodsFor: 'testing' stamp: 'di 1/30/98 13:29'!
stepTime
	^ 100! !

!EnvelopeEditorMorph methodsFor: 'scaling' stamp: 'di 1/31/98 01:10'!
valueFromY: y
	"The convention is that envelope values are between 0.0 and 1.0"
	| value |
	value := (graphArea bottom - y) asFloat / (graphArea height).
	envelope updateSelector = #ratio: ifTrue:
		["Ratio gets gridded by denominator"
		^ (value * envelope scale * denominator) rounded asFloat / denominator / envelope scale].
	^ value! !

!EnvelopeEditorMorph methodsFor: 'rounding' stamp: 'di 7/14/2000 11:13'!
wantsRoundedCorners
	^ Preferences roundedWindowCorners or: [super wantsRoundedCorners]! !

!EnvelopeEditorMorph methodsFor: 'scaling' stamp: 'di 1/26/98 14:09'!
xFromMs: ms
	^ graphArea left + (ms//10*pixPerTick)! !

!EnvelopeEditorMorph methodsFor: 'scaling' stamp: 'di 1/27/98 00:23'!
yFromValue: val
	"The convention is that envelope values are between 0.0 and 1.0"
	^ graphArea bottom - (val* (graphArea height))! !

RectangleMorph subclass: #GraphMorph
	instanceVariableNames: 'data dataColor cursor cursorColor cursorColorAtZeroCrossings startIndex minVal maxVal cachedForm hasChanged'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!GraphMorph commentStamp: 'StephaneDucasse 8/20/2009 16:23' prior: 0!
I display a graph of numbers, normalized so the full range of values just fits my height. I support a movable cursor that can be dragged with the mouse.

Implementation notes: Some operations on me may be done at sound sampling rates (e.g. 11-44 thousand times/second). To allow such high bandwidth application, certain operations that change my appearance do not immediately report a damage rectangle. Instead, a flag is set indicating that my display needs to refreshed and a step method reports the damage rectangle if that flag is set. Also, I cache a bitmap of my graph to allow the cursor to be moved without redrawing the graph.

	GraphMorph new openInWorld !

!GraphMorph class methodsFor: 'parts bin' stamp: 'sw 8/2/2001 12:50'!
descriptionForPartsBin
	^ self partName:	'Graph'
		categories:		#('Useful')
		documentation:	'A graph of numbers, normalized so the full range of values just fits my height.  I support a movable cursor that can be dragged with the mouse.'! !

!GraphMorph methodsFor: 'commands' stamp: 'marcus.denker 8/17/2008 21:25'!
appendValue: aPointOrNumber

	| newVal |
	data := data asOrderedCollection.
	newVal := self asNumber: aPointOrNumber.
	data addLast: newVal.
	newVal < minVal ifTrue: [minVal := newVal].
	newVal > maxVal ifTrue: [maxVal := newVal].
	self cursor: data size.
	self flushCachedForm.
! !

!GraphMorph methodsFor: 'commands' stamp: 'jm 7/30/1998 12:17'!
centerCursor
	"Scroll so that the cursor is as close as possible to the center of my window."

	| w |
	w := self width - (2 * borderWidth).
	self startIndex: ((cursor - (w // 2)) max: 1).
! !

!GraphMorph methodsFor: 'commands' stamp: ''!
clear

	self startIndex: 1.
	self cursor: 1.
	self data: OrderedCollection new.
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
color: aColor

	super color: aColor.
	self flushCachedForm.
! !

!GraphMorph methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:46'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	hasChanged ifNil: [hasChanged := false].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!GraphMorph methodsFor: 'e-toy support' stamp: ''!
cursor

	^ cursor
! !

!GraphMorph methodsFor: 'e-toy support' stamp: 'jm 6/17/1999 21:41'!
cursor: aNumber

	| truncP |
	cursor ~= aNumber ifTrue:  [
		cursor := aNumber.
		truncP := aNumber truncated.
		truncP > data size ifTrue: [cursor := data size].
		truncP < 0 ifTrue: [cursor := 1].
		self keepIndexInView: truncP.
		hasChanged := true].
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
cursorAtEnd

	^ cursor truncated >= data size
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
cursorColor

	^ cursorColor
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
cursorColor: aColor

	cursorColor := aColor.
	self flushCachedForm.
! !

!GraphMorph methodsFor: 'accessing' stamp: 'jm 7/8/1998 20:32'!
cursorColorAtZeroCrossing

	^ cursorColorAtZeroCrossings
! !

!GraphMorph methodsFor: 'accessing' stamp: 'jm 7/8/1998 20:32'!
cursorColorAtZeroCrossings: aColor

	cursorColorAtZeroCrossings := aColor.
	self flushCachedForm.
! !

!GraphMorph methodsFor: 'accessing' stamp: 'jm 6/17/1999 21:43'!
cursorWrapped: aNumber

	| sz |
	cursor ~= aNumber ifTrue: [
		cursor := aNumber.
		sz := data size.
		sz = 0
			ifTrue: [cursor := 1]
			ifFalse: [
				((cursor >= (sz + 1)) or: [cursor < 0]) ifTrue: [
					cursor := cursor - ((cursor // sz) * sz)].
				cursor < 1 ifTrue: [cursor := sz + cursor]].
		"assert: 1 <= cursor < data size + 1"
		hasChanged := true].
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
data

	^ data
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
data: aCollection

	data := aCollection.
	maxVal := minVal := 0.
	data do: [:x |
		x < minVal ifTrue: [minVal := x].
		x > maxVal ifTrue: [maxVal := x]].

	self flushCachedForm.
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
dataColor

	^ dataColor
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
dataColor: aColor

	dataColor := aColor.
	self flushCachedForm.
! !

!GraphMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:27'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color
		r: 0.8
		g: 0.8
		b: 0.6! !

!GraphMorph methodsFor: 'private' stamp: 'jm 6/17/1999 21:36'!
drawCursorOn: aCanvas

	| ptr x r c |
	ptr := (cursor asInteger max: 1) min: data size.
	c := cursorColor.
	((ptr > 1) and: [ptr < data size]) ifTrue: [
		(data at: ptr) sign ~= (data at: ptr + 1) sign
			ifTrue: [c := cursorColorAtZeroCrossings]].
	r := self innerBounds.
	x := r left + ptr - startIndex.
	((x >= r left) and: [x <= r right]) ifTrue: [
		aCanvas fillRectangle: (x@r top corner: x + 1@r bottom) color: c].
! !

!GraphMorph methodsFor: 'private' stamp: ''!
drawDataOn: aCanvas

	| yScale baseLine x start end value left top bottom right |
	super drawOn: aCanvas.

	data isEmpty ifTrue: [^ self].
	maxVal = minVal ifTrue: [
		yScale := 1.
	] ifFalse: [
		yScale := (bounds height - (2 * borderWidth)) asFloat / (maxVal - minVal)].
	baseLine := bounds bottom - borderWidth + (minVal * yScale) truncated.
	left := top := 0. right := 10. bottom := 0.
	x := bounds left + borderWidth.
	start := (startIndex asInteger max: 1) min: data size.
	end := (start + bounds width) min: data size.
	start to: end do: [:i |
		left := x truncated. right := x + 1.
		right > (bounds right - borderWidth) ifTrue: [^ self].
		value := (data at: i) asFloat.
		value >= 0.0 ifTrue: [
			top := baseLine - (yScale * value) truncated.
			bottom := baseLine.
		] ifFalse: [
			top := baseLine.
			bottom := baseLine - (yScale * value) truncated].
		aCanvas fillRectangle: (left@top corner: right@bottom) color: dataColor.
		x := x + 1].
! !

!GraphMorph methodsFor: 'drawing' stamp: 'dgd 2/22/2003 14:39'!
drawOn: aCanvas 
	| c |
	cachedForm isNil 
		ifTrue: 
			[c := Display defaultCanvasClass extent: bounds extent.
			c translateBy: bounds origin negated
				during: [:tempCanvas | self drawDataOn: tempCanvas].
			cachedForm := c form].
	aCanvas 
		cache: bounds
		using: cachedForm
		during: [:cachingCanvas | self drawDataOn: cachingCanvas].
	self drawCursorOn: aCanvas! !

!GraphMorph methodsFor: 'private' stamp: 'jm 6/17/1999 20:10'!
flushCachedForm

	cachedForm := nil.
	hasChanged := true.
! !

!GraphMorph methodsFor: 'event handling' stamp: 'jm 10/18/97 11:32'!
handlesMouseDown: evt

	evt shiftPressed
		ifTrue: [^ super handlesMouseDown: evt]
		ifFalse: [^ true].
! !

!GraphMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:28'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self extent: 365 @ 80.

	dataColor := Color darkGray.
	cursor := 1.0.
	"may be fractional"
	cursorColor := Color red.
	cursorColorAtZeroCrossings := Color red.
	startIndex := 1.
	hasChanged := false.
	self
		data: ((0 to: 360 - 1)
				collect: [:x | (100.0 * x degreesToRadians sin) asInteger])! !

!GraphMorph methodsFor: 'accessing' stamp: 'dgd 2/22/2003 14:40'!
interpolatedValueAtCursor
	| sz prev frac next |
	data isEmpty ifTrue: [^0].
	sz := data size.
	cursor < 0 ifTrue: [^data first].	"just to be safe, though cursor shouldn't be negative"
	prev := cursor truncated.
	frac := cursor - prev.
	prev < 1 ifTrue: [prev := sz].
	prev > sz ifTrue: [prev := 1].
	"assert: 1 <= prev <= sz"
	frac = 0 ifTrue: [^data at: prev].	"no interpolation needed"

	"interpolate"
	next := prev = sz ifTrue: [1] ifFalse: [prev + 1].
	^(1.0 - frac) * (data at: prev) + (frac * (data at: next))! !

!GraphMorph methodsFor: 'private' stamp: 'jm 4/21/1999 11:30'!
keepIndexInView: index

	| w newStart |
	w := bounds width - (2 * borderWidth).
	index < startIndex ifTrue: [
		newStart := index - w + 1.
		^ self startIndex: (newStart max: 1)].
	index > (startIndex + w) ifTrue: [
		^ self startIndex: (index min: data size)].
! !

!GraphMorph methodsFor: 'accessing' stamp: 'dgd 2/22/2003 14:40'!
lastValue
	data isEmpty ifTrue: [^0].
	^data last! !

!GraphMorph methodsFor: 'accessing' stamp: 'jm 4/21/1999 11:25'!
lastValue: aNumber

	self appendValue: aNumber.
! !

!GraphMorph methodsFor: 'layout' stamp: ''!
layoutChanged

	super layoutChanged.
	cachedForm := nil.
! !

!GraphMorph methodsFor: 'commands' stamp: 'ads 7/31/2003 11:11'!
loadSineWave

	self loadSoundData: SoundBuffer sineTable.
! !

!GraphMorph methodsFor: 'commands' stamp: 'jm 4/22/1999 14:17'!
loadSound: aSound

	self loadSoundData: aSound samples.
! !

!GraphMorph methodsFor: 'commands' stamp: 'jm 4/22/1999 14:23'!
loadSoundData: aCollection

	| scale absV newData |
	scale := 0.
	aCollection do: [:v | (absV := v abs) > scale ifTrue: [scale := absV]].
	scale := 100.0 / scale.
	newData := OrderedCollection new: aCollection size.
	1 to: aCollection size do: [:i | newData addLast: (scale * (aCollection at: i))].

	self data: newData.
	self startIndex: 1.
	self cursor: 1.
! !

!GraphMorph methodsFor: 'event handling' stamp: 'jm 10/18/97 11:52'!
mouseMove: evt

	| x w |
	x := evt cursorPoint x - (bounds left + borderWidth).
	w := self width - (2 * borderWidth).

	self changed.
	x < 0 ifTrue: [
		cursor := startIndex + (3 * x).
		cursor := (cursor max: 1) min: data size.
		^ self startIndex: cursor].
	x > w ifTrue: [
		cursor := startIndex + w + (3 * (x - w)).
		cursor := (cursor max: 1) min: data size.
		^ self startIndex: cursor - w].

	cursor := ((startIndex + x) max: 1) min: data size.
! !

!GraphMorph methodsFor: 'commands' stamp: 'gk 2/23/2004 21:08'!
playOnce

	| scale absV scaledData |
	data isEmpty ifTrue: [^ self].  "nothing to play"
	scale := 1.
	data do: [:v | (absV := v abs) > scale ifTrue: [scale := absV]].
	scale := 32767.0 / scale.
	scaledData := SoundBuffer newMonoSampleCount: data size.
	1 to: data size do: [:i | scaledData at: i put: (scale * (data at: i)) truncated].
	SoundService default playSampledSound: scaledData rate: 11025.
! !

!GraphMorph methodsFor: 'commands' stamp: ''!
reverse

	data := data reversed.
	self flushCachedForm.
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
startIndex

	^ startIndex
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
startIndex: aNumber

	startIndex ~= aNumber ifTrue:  [
		startIndex := aNumber asInteger.
		self flushCachedForm].
! !

!GraphMorph methodsFor: 'stepping and presenter' stamp: 'dgd 2/22/2003 14:40'!
step
	"Make a deferred damage rectangle if I've changed. This allows applications to call methods that invalidate my display at high-bandwidth without paying the cost of doing the damage reporting on ever call; they can merely set hasChanged to true."

	super step.
	hasChanged isNil ifTrue: [hasChanged := false].
	hasChanged 
		ifTrue: 
			[self changed.
			hasChanged := false]! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
valueAtCursor

	data isEmpty ifTrue: [^ 0].
	^ data at: ((cursor truncated max: 1) min: data size)
! !

!GraphMorph methodsFor: 'accessing' stamp: ''!
valueAtCursor: aPointOrNumber

	data isEmpty ifTrue: [^ 0].
	data
		at: ((cursor truncated max: 1) min: data size)
		put: (self asNumber: aPointOrNumber).
	self flushCachedForm.
! !

RectangleMorph subclass: #InterimSoundMorph
	instanceVariableNames: 'graphic sound'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!InterimSoundMorph methodsFor: 'as yet unclassified' stamp: 'di 2/17/2000 20:30'!
addGraphic

	graphic := SketchMorph withForm: self speakerGraphic.
	graphic position: bounds center - (graphic extent // 2).
	self addMorph: graphic.
! !

!InterimSoundMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:38'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!InterimSoundMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:27'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color
		r: 0
		g: 0.8
		b: 0.6! !

!InterimSoundMorph methodsFor: 'event handling' stamp: 'jm 9/18/97 15:26'!
handlesMouseDown: evt

	(graphic containsPoint: evt cursorPoint)
		ifTrue: [^ true]
		ifFalse: [^ super handlesMouseDown: evt].
! !

!InterimSoundMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:57'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	self extent: 30 @ 30.
	self addGraphic.
	sound := PluckedSound
				pitch: 880.0
				dur: 2.0
				loudness: 0.5! !

!InterimSoundMorph methodsFor: 'event handling' stamp: 'jm 9/18/97 18:56'!
mouseDown: evt

	(graphic containsPoint: evt cursorPoint)
		ifTrue: [sound copy play]
		ifFalse: [super mouseDown: evt].
! !

!InterimSoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 9/18/97 15:02'!
sound

	^ sound
! !

!InterimSoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 9/18/97 15:01'!
sound: aSound

	sound := aSound.
! !

!InterimSoundMorph methodsFor: 'as yet unclassified' stamp: 'jm 9/18/97 15:20'!
speakerGraphic

	^ Form
		extent: 19@18
		depth: 8
		fromArray: #(0 0 1493172224 2816 0 0 0 1493172224 11 0 0 138 1493172224 184549376 184549376 0 35509 2315255808 720896 720896 0 9090522 2315255808 2816 720896 0 2327173887 2315255819 2816 720896 138 3051028442 2315255819 2816 2816 1505080590 4294957786 2315255808 184549387 2816 3053453311 4292532917 1493172224 184549387 2816 1505080714 3048584629 1493172224 184549387 2816 9079434 3048584629 1493172224 184549387 2816 138 2327164341 1493172235 2816 2816 0 2324346293 1493172235 2816 720896 0 9079477 1493172224 2816 720896 0 35466 1493172224 720896 720896 0 138 0 184549376 184549376 0 0 0 11 0 0 0 0 2816 0)
		offset: 0@0
! !

RectangleMorph subclass: #PianoKeyboardMorph
	instanceVariableNames: 'whiteKeyColor blackKeyColor playingKeyColor nOctaves target noteOnSelector noteOffSelector soundPrototype soundPlaying'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

PianoKeyboardMorph subclass: #KeyboardMorphForInput
	instanceVariableNames: 'pianoRoll duration durationModifier articulation buildingChord insertMode prevSelection startOfNextNote'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!KeyboardMorphForInput commentStamp: '<historical>' prior: 0!
This class adds state and controls to the basic PianoKeyboardMorph so that notes of reliable duration can be keyed into a score without the need for a real keyboard.

To try this out, execute...

	| n score | n _ 3.
	score _ (MIDIScore new tracks: ((1 to: n) collect: [:i | Array new]);
		trackInfo: ((1 to: n) collect: [:i | 'Instrument' , i printString]);
		tempoMap: nil; ticksPerQuarterNote: 96).
	ScorePlayerMorph openOn: score title: 'empty score'

Then open a pianoRoll and, from that, open a keyboard.  The rule is that the keyboard will append after the current selection.  If the current selection is muted or nil, then input will go to the end of the first non-muted track.!

!KeyboardMorphForInput methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
addRecordingControls
	| button switch playRow durRow articRow modRow |

	"Add chord, rest and delete buttons"
	playRow := AlignmentMorph newRow.
	playRow color: color; borderWidth: 0; layoutInset: 0.
	playRow hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	playRow addMorphBack: (switch label: 'chord' translated; actionSelector: #buildChord:).
	button := SimpleButtonMorph new target: self;
		borderColor: #raised; borderWidth: 2; color: color.
	playRow addMorphBack: (button label: '          rest          ' translated; actionSelector: #emitRest).
	button := SimpleButtonMorph new target: self;
		borderColor: #raised; borderWidth: 2; color: color.
	playRow addMorphBack: (button label: 'del' translated; actionSelector: #deleteNotes).
	self addMorph: playRow.
	playRow align: playRow fullBounds topCenter
			with: self fullBounds bottomCenter.

	"Add note duration buttons"
	durRow := AlignmentMorph newRow.
	durRow color: color; borderWidth: 0; layoutInset: 0.
	durRow hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	durRow addMorphBack: (switch label: 'whole' translated;
				actionSelector: #duration:onOff:; arguments: #(1)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	durRow addMorphBack: (switch label: 'half' translated;
				actionSelector: #duration:onOff:; arguments: #(2)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	durRow addMorphBack: (switch label: 'quarter' translated;
				actionSelector: #duration:onOff:; arguments: #(4)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	durRow addMorphBack: (switch label: 'eighth' translated;
				actionSelector: #duration:onOff:; arguments: #(8)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	durRow addMorphBack: (switch label: 'sixteenth' translated;
				actionSelector: #duration:onOff:; arguments: #(16)).
	self addMorph: durRow.
	durRow align: durRow fullBounds topCenter
			with: playRow fullBounds bottomCenter.

	"Add note duration modifier buttons"
	modRow := AlignmentMorph newRow.
	modRow color: color; borderWidth: 0; layoutInset: 0.
	modRow hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	modRow addMorphBack: (switch label: 'dotted' translated;
				actionSelector: #durMod:onOff:; arguments: #(dotted)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	modRow addMorphBack: (switch label: 'normal' translated;
				actionSelector: #durMod:onOff:; arguments: #(normal)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	modRow addMorphBack: (switch label: 'triplets' translated;
				actionSelector: #durMod:onOff:; arguments: #(triplets)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	modRow addMorphBack: (switch label: 'quints' translated;
				actionSelector: #durMod:onOff:; arguments: #(quints)).
	self addMorph: modRow.
	modRow align: modRow fullBounds topCenter
			with: durRow fullBounds bottomCenter.

	"Add articulation buttons"
	articRow := AlignmentMorph newRow.
	articRow color: color; borderWidth: 0; layoutInset: 0.
	articRow hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	articRow addMorphBack: (switch label: 'legato' translated;
				actionSelector: #articulation:onOff:; arguments: #(legato)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	articRow addMorphBack: (switch label: 'normal' translated;
				actionSelector: #articulation:onOff:; arguments: #(normal)).
	switch := SimpleSwitchMorph new target: self; borderWidth: 2;
		offColor: color; onColor: (Color r: 1.0 g: 0.6 b: 0.6); setSwitchState: false.
	articRow addMorphBack: (switch label: 'staccato' translated;
				actionSelector: #articulation:onOff:; arguments: #(staccato)).
	self addMorph: articRow.
	articRow align: articRow fullBounds topCenter
			with: modRow fullBounds bottomCenter.

	self bounds: (self fullBounds expandBy: (0@0 extent: 0@borderWidth))
! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'stephaneducasse 2/4/2006 20:41'!
articulation: artic onOff: ignored    "artic = eg, #legato, #normal, #staccato."
	"Set the articulation of notes to be emitted when a key is pressed."

	self allMorphsDo:
		[:m | ((m isMemberOf: SimpleSwitchMorph)
				and: [m actionSelector == #articulation:onOff:])
				ifTrue: [m setSwitchState: m arguments first == artic]].
	articulation := artic! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'di 6/20/1999 00:20'!
backspaceNote

	self deleteNotes! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'di 6/18/1999 15:50'!
buildChord: onOff! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'di 6/20/1999 00:20'!
deleteNotes

	pianoRoll deleteSelection! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'stephaneducasse 2/4/2006 20:41'!
durMod: durMod onOff: ignored    "durMod = eg, #dotted, #normal, #triplets, #quints"
	"Set the duration of notes to be emitted when a key is pressed."

	self allMorphsDo:
		[:m | ((m isMemberOf: SimpleSwitchMorph)
				and: [m actionSelector == #durMod:onOff:])
				ifTrue: [m setSwitchState: m arguments first = durMod]].
	durationModifier := durMod! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'stephaneducasse 2/4/2006 20:41'!
duration: denom onOff: ignored    "denom = eg, 1, 2, 4, 8, 16"
	"Set the duration of notes to be emitted when a key is pressed."

	self allMorphsDo:
		[:m | ((m isMemberOf: SimpleSwitchMorph)
				and: [m actionSelector == #duration:onOff:])
				ifTrue: [m setSwitchState: m arguments first = denom]].
	duration := denom.
	self durMod: #normal onOff: true! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'stephaneducasse 2/4/2006 20:41'!
emitRest

	| sel noteEvent |

	"All this selection logic should be shared with mouseDown..."
	(sel := pianoRoll selection) ifNil: [^ self].
	insertMode ifTrue:
		[sel := pianoRoll selectionForInsertion.
		insertMode := false].
	sel = prevSelection ifFalse:
		["This is a new selection -- need to determine start time"
		sel third = 0
			ifTrue: [startOfNextNote := 0]
			ifFalse: [startOfNextNote := ((pianoRoll score tracks at: sel first)
										at: sel third) endTime.
					startOfNextNote := startOfNextNote + self fullDuration - 1
										truncateTo: self fullDuration]].
	noteEvent := NoteEvent new time: startOfNextNote; duration: self noteDuration;
			key: -1 "my flag for rest" velocity: self velocity channel: 1.
	pianoRoll appendEvent: noteEvent fullDuration: self fullDuration.
	soundPlaying ifNotNil: [soundPlaying stopGracefully].
	prevSelection := pianoRoll selection.
	startOfNextNote := startOfNextNote + self fullDuration.! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'stephaneducasse 2/4/2006 20:41'!
fullDuration

	| num denom |
	num := denom := 1.
	durationModifier == #dotted ifTrue: [num := 3.  denom := 2].
	durationModifier == #triplets ifTrue: [num := 2.  denom := 3].
	durationModifier == #quints ifTrue: [num := 2.  denom := 5].
	^ pianoRoll score ticksPerQuarterNote * 4 * num // duration // denom! !

!KeyboardMorphForInput methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	buildingChord := false.
	self addRecordingControls.
	self duration: 4 onOff: true.
	self durMod: #normal onOff: true.
	self articulation: #normal onOff: true.
	insertMode := false! !

!KeyboardMorphForInput methodsFor: 'simple keyboard' stamp: 'stephaneducasse 2/4/2006 20:41'!
mouseDownPitch: midiKey event: event noteMorph: keyMorph

	| sel noteEvent |
	event hand hasSubmorphs ifTrue: [^ self  "no response if drag something over me"].
	keyMorph color: playingKeyColor.
	(sel := pianoRoll selection) ifNil: [^ self].
	insertMode ifTrue:
		[sel := pianoRoll selectionForInsertion.
		insertMode := false].
	sel = prevSelection ifFalse:
		["This is a new selection -- need to determine start time"
		sel third = 0
			ifTrue: [startOfNextNote := 0]
			ifFalse: [startOfNextNote := ((pianoRoll score tracks at: sel first)
										at: sel third) endTime.
					startOfNextNote := startOfNextNote + self fullDuration - 1
										truncateTo: self fullDuration]].
	noteEvent := NoteEvent new time: startOfNextNote; duration: self noteDuration;
			key: midiKey + 23 velocity: self velocity channel: 1.
	pianoRoll appendEvent: noteEvent fullDuration: self fullDuration.
	soundPlaying ifNotNil: [soundPlaying stopGracefully].
	(soundPlaying := self soundForEvent: noteEvent inTrack: sel first) play.
	prevSelection := pianoRoll selection.
	startOfNextNote := startOfNextNote + self fullDuration.! !

!KeyboardMorphForInput methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:28'!
mouseUpPitch: pitch event: event noteMorph: noteMorph
	noteMorph color: ((#(0 1 3 5 6 8 10) includes: pitch\\12)
					ifTrue: [whiteKeyColor]
					ifFalse: [blackKeyColor]).
! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'di 6/20/1999 00:27'!
noteDuration

	articulation == #staccato ifTrue: [^ (self fullDuration * 0.65) asInteger].
	articulation == #normal ifTrue: [^ (self fullDuration * 0.8) asInteger].
	articulation == #legato ifTrue: [^ (self fullDuration * 0.95) asInteger].
! !

!KeyboardMorphForInput methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
pianoRoll: prMorph

	pianoRoll := prMorph! !

!KeyboardMorphForInput methodsFor: 'events' stamp: 'stephaneducasse 2/4/2006 20:41'!
soundForEvent: noteEvent inTrack: trackIndex

	| sound player |
	player := pianoRoll scorePlayer.
	sound := MixedSound new.
	sound add: ((player instrumentForTrack: trackIndex)
					soundForMidiKey: noteEvent midiKey
					dur: noteEvent duration / (pianoRoll scorePlayer ticksForMSecs: 1000)
					loudness: (noteEvent velocity asFloat / 127.0))
			pan: (player panForTrack: trackIndex)
			volume: player overallVolume *
						(player volumeForTrack: trackIndex).
	^ sound
! !

!KeyboardMorphForInput methodsFor: 'note controls' stamp: 'di 6/19/1999 23:13'!
velocity

	^ 80  "Later put a slider on the keyboard control"! !

PianoKeyboardMorph subclass: #MIDIPianoKeyboardMorph
	instanceVariableNames: 'midiPort channel velocity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!MIDIPianoKeyboardMorph commentStamp: '<historical>' prior: 0!
I implement a piano keyboard that can be played with the mouse. I can output to a MIDI port, if MIDI is supported on this platform. I can also spawn controllers for other MIDI parameters, such as pitch bend.
!

!MIDIPianoKeyboardMorph methodsFor: 'menus' stamp: 'dgd 8/30/2003 21:50'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	midiPort
		ifNil: [aCustomMenu add: 'play via MIDI' translated action: #openMIDIPort]
		ifNotNil: [
			aCustomMenu add: 'play via built in synth' translated action: #closeMIDIPort.
			aCustomMenu add: 'new MIDI controller' translated action: #makeMIDIController:].
! !

!MIDIPianoKeyboardMorph methodsFor: 'as yet unclassified' stamp: 'jm 10/12/1998 15:57'!
closeMIDIPort

	midiPort := nil.
! !

!MIDIPianoKeyboardMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:55'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	SimpleMIDIPort midiIsSupported
		ifTrue: [midiPort := SimpleMIDIPort openDefault].
	channel := 1.
	velocity := 100! !

!MIDIPianoKeyboardMorph methodsFor: 'as yet unclassified' stamp: 'jm 10/12/1998 18:11'!
makeMIDIController: evt

	self world activeHand attachMorph:
		(MIDIControllerMorph new midiPort: midiPort).
! !

!MIDIPianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:40'!
mouseDownPitch: midiKey event: event noteMorph: noteMorph

	midiPort ifNil: [^ super mouseDownPitch: midiKey-1 event: event noteMorph: noteMorph].
	noteMorph color: playingKeyColor.
	soundPlaying
		ifNil: [midiPort ensureOpen]
		ifNotNil: [self turnOffNote].
	self turnOnNote: midiKey + 23.
! !

!MIDIPianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:41'!
mouseUpPitch: midiKey event: event noteMorph: noteMorph

	midiPort ifNil: [
		^ super mouseUpPitch: midiKey event: event noteMorph: noteMorph].

	noteMorph color:
		((#(0 1 3 5 6 8 10) includes: midiKey \\ 12)
			ifTrue: [whiteKeyColor]
			ifFalse: [blackKeyColor]).
	soundPlaying ifNotNil: [self turnOffNote].
! !

!MIDIPianoKeyboardMorph methodsFor: 'as yet unclassified' stamp: 'jm 10/12/1998 18:00'!
openMIDIPort

	| portNum |
	portNum := SimpleMIDIPort outputPortNumFromUser.
	portNum ifNil: [^ self].
	midiPort := SimpleMIDIPort openOnPortNumber: portNum.
! !

!MIDIPianoKeyboardMorph methodsFor: 'as yet unclassified' stamp: 'jm 10/12/1998 15:44'!
turnOffNote

	midiPort notNil & soundPlaying notNil ifTrue: [
		soundPlaying isInteger ifTrue: [
			midiPort midiCmd: 16r90 channel: channel byte: soundPlaying byte: 0]].
	soundPlaying := nil.
! !

!MIDIPianoKeyboardMorph methodsFor: 'as yet unclassified' stamp: 'jm 10/12/1998 16:25'!
turnOnNote: midiKey

	midiPort midiCmd: 16r90 channel: channel byte: midiKey byte: velocity.
	soundPlaying := midiKey.
! !

!PianoKeyboardMorph class methodsFor: 'parts bin' stamp: 'sw 8/2/2001 14:52'!
descriptionForPartsBin
	^ self partName:	'PianoKeyboard'
		categories:		#('Multimedia')
		documentation:	'A piano keyboard'! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:30'!
buildKeyboard
	| wtWid bkWid keyRect octavePt nWhite nBlack |
	self removeAllMorphs.
	wtWid := 8. bkWid := 5.
	self extent: 10@10.
	1 to: nOctaves+1 do:
		[:i | i <= nOctaves ifTrue: [nWhite := 7.  nBlack := 5]
						ifFalse: [nWhite := 1.  nBlack := 0 "High C"].
		octavePt := self innerBounds topLeft + ((7*wtWid*(i-1)-1)@-1).
		1 to: nWhite do:
			[:j | keyRect := octavePt + (j-1*wtWid@0) extent: (wtWid+1)@36.
			self addMorph: ((RectangleMorph newBounds: keyRect color: whiteKeyColor)
								borderWidth: 1;
				on: #mouseDown send: #mouseDownPitch:event:noteMorph: to: self
								withValue: i-1*12 + (#(1 3 5 6 8 10 12) at: j))].
		1 to: nBlack do:
			[:j | keyRect := octavePt + ((#(6 15 29 38 47) at: j)@1) extent: bkWid@21.
			self addMorph: ((Morph newBounds: keyRect color: blackKeyColor)
				on: #mouseDown send: #mouseDownPitch:event:noteMorph: to: self
								withValue: i-1*12 + (#(2 4 7 9 11) at: j))]].
	self submorphsDo:
		[:m | m on: #mouseMove send: #mouseMovePitch:event:noteMorph: to: self;
				on: #mouseUp send: #mouseUpPitch:event:noteMorph: to: self;
				on: #mouseEnterDragging send: #mouseDownPitch:event:noteMorph: to: self;
				on: #mouseLeaveDragging send: #mouseUpPitch:event:noteMorph: to: self].
	self extent: (self fullBounds extent + borderWidth - 1)! !

!PianoKeyboardMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:29'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color veryLightGray! !

!PianoKeyboardMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:56'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	whiteKeyColor := Color gray: 0.95.
	blackKeyColor := Color black.
	playingKeyColor := Color red.
	nOctaves := 6.
	self buildKeyboard.
	soundPrototype := FMSound brass1 duration: 9.9! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/18/2001 17:27'!
mouseDownEvent: arg1 noteMorph: arg2 pitch: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self mouseDownPitch: arg1 event: arg2 noteMorph: arg3! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:27'!
mouseDownPitch: midiKey event: event noteMorph: noteMorph
	| pitch |
	event hand hasSubmorphs ifTrue: [^ self  "no response if drag something over me"].
	event hand mouseFocus ifNil:
		["If dragged into me, then establish focus so I'll see moves"
		event hand newMouseFocus: noteMorph event: event].
	noteMorph color: playingKeyColor.
	pitch := AbstractSound pitchForMIDIKey: midiKey + 23.
	soundPlaying ifNotNil: [soundPlaying stopGracefully].
	soundPlaying := soundPrototype soundForPitch: pitch dur: 100.0 loudness: 0.3.
	SoundPlayer resumePlaying: soundPlaying quickStart: true.
! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/18/2001 17:27'!
mouseMoveEvent: arg1 noteMorph: arg2 pitch: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self mouseMovePitch: arg1 event: arg2 noteMorph: arg3! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:28'!
mouseMovePitch: pitch event: event noteMorph: noteMorph

	(noteMorph containsPoint: event cursorPoint) ifFalse:
		["If drag out of me, zap focus so other morphs can see drag in."
		event hand releaseMouseFocus: noteMorph]
! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/18/2001 17:28'!
mouseUpEvent: arg1 noteMorph: arg2 pitch: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self mouseUpPitch: arg1 event: arg2 noteMorph: arg3! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'ar 3/17/2001 14:29'!
mouseUpPitch: pitch event: event noteMorph: noteMorph

	noteMorph color: ((#(0 1 3 5 6 8 10) includes: pitch\\12)
					ifTrue: [whiteKeyColor]
					ifFalse: [blackKeyColor]).
	soundPlaying ifNotNil: [soundPlaying stopGracefully].
! !

!PianoKeyboardMorph methodsFor: 'simple keyboard' stamp: 'di 9/4/1998 12:57'!
soundPrototype: aSound
	soundPrototype := aSound! !

RectangleMorph subclass: #PianoRollScoreMorph
	instanceVariableNames: 'scorePlayer score colorForTrack lowestNote leftEdgeTime timeScale indexInTrack lastUpdateTick lastMutedState cursor selection timeSignature beatsPerMeasure notePerBeat showMeasureLines showBeatLines soundsPlaying soundsPlayingMorph movieClipPlayer'
	classVariableNames: 'NotePasteBuffer'
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!PianoRollScoreMorph commentStamp: '<historical>' prior: 0!
A PianoRollScoreMorph displays a score such as a MIDIScore, and will scroll through it tracking the progress of a ScorePlayerMorph (from which it is usually spawned).

timeScale is in pixels per score tick.

Currently the ambient track (for synchronizing thumbnails, eg) is treated specially here and in the score.  This should be cleaned up by adding a trackType or something like it in the score.!

!PianoRollScoreMorph methodsFor: 'layout' stamp: 'RAA 12/11/2000 23:00'!
acceptDroppingMorph: aMorph event: evt
	"In addition to placing this morph in the pianoRoll, add a corresponding
	event to the score so that it will always appear when played, in addition
	to possibly triggering other actions"

	aMorph justDroppedIntoPianoRoll: self event: evt.
	super acceptDroppingMorph: aMorph event: evt.

! !

!PianoRollScoreMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:56'!
addCustomMenuItems: aMenu hand: aHandMorph

	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu add: 'expand time' translated action: #expandTime.
	aMenu add: 'contract time' translated action: #contractTime.
	aMenu addLine.
	aMenu add: 'add movie clip player' translated action: #addMovieClipPlayer.
	(self valueOfProperty: #dragNDropEnabled) == true
		ifTrue: [aMenu add: 'close drag and drop' translated action: #disableDragNDrop]
		ifFalse: [aMenu add: 'open drag and drop' translated action: #enableDragNDrop].
! !

!PianoRollScoreMorph methodsFor: 'menu' stamp: 'di 6/18/1999 15:55'!
addKeyboard

	(KeyboardMorphForInput new pianoRoll: self) openInWorld! !

!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'stephaneducasse 2/4/2006 20:41'!
addNotes
	"Recompute the set of morphs that should be visible at the current scroll position."

	| visibleMorphs rightEdge topEdge trackColor i done n nLeft nTop nRight rightEdgeTime |
	visibleMorphs := OrderedCollection new: 500.
	rightEdge := self right - borderWidth.
	rightEdgeTime := self timeForX: rightEdge.
	topEdge := self top + borderWidth + 1.

	"Add ambient morphs first (they will be front-most)"
	score eventMorphsWithTimeDo:
		[:m :t | m addMorphsTo: visibleMorphs pianoRoll: self eventTime: t
					betweenTime: leftEdgeTime and: rightEdgeTime].

	"Then add note morphs"
	score tracks withIndexDo:
		[:track :trackIndex |
		trackColor := colorForTrack at: trackIndex.
		i := indexInTrack at: trackIndex.
		done := scorePlayer mutedForTrack: trackIndex.
		[done | (i > track size)] whileFalse: [
			n := track at: i.
			(n isNoteEvent and: [n midiKey >= lowestNote]) ifTrue: [
				n time > rightEdgeTime
					ifTrue: [done := true]
					ifFalse: [
						nLeft := self xForTime: n time.
						nTop := (self yForMidiKey: n midiKey) - 1.
						nTop > topEdge ifTrue: [
							nRight := nLeft + (n duration * timeScale) truncated - 1.
							visibleMorphs add:
								((PianoRollNoteMorph
									newBounds: (nLeft@nTop corner: nRight@(nTop + 3))
									color: trackColor)
									trackIndex: trackIndex indexInTrack: i)]]].
			i := i + 1].
			(selection notNil
				and: [trackIndex = selection first
				and: [i >= selection second and: [(indexInTrack at: trackIndex) <= selection third]]])
				ifTrue: [visibleMorphs do:
						[:vm | (vm isKindOf: PianoRollNoteMorph) ifTrue: [vm selectFrom: selection]]]].

	"Add the cursor morph in front of all notes; height and position are set later."
	cursor ifNil: [cursor := Morph newBounds: (self topLeft extent: 1@1) color: Color red].
	visibleMorphs addFirst: cursor.

	self changed.
	self removeAllMorphs.
	self addAllMorphs: visibleMorphs.
! !

!PianoRollScoreMorph methodsFor: 'editing' stamp: 'stephaneducasse 2/4/2006 20:41'!
appendEvent: noteEvent fullDuration: fullDuration 

	| sel x |
	score appendEvent: noteEvent fullDuration: fullDuration at: (sel := self selection).
	noteEvent midiKey = -1 ifFalse:  "Unless it is a rest..."
		["Advance the selection to the note just entered"
		selection := Array with: sel first with: sel third + 1 with: sel third + 1].

	"This is all horribly inefficient..."
	scorePlayer updateDuration.
	(x := self xForTime: noteEvent endTime) > (self right - 30) ifTrue:
		[self autoScrollForX: x + (30 + self width // 4)].
	self updateLowestNote.
	self rebuildFromScore! !

!PianoRollScoreMorph methodsFor: 'scrolling' stamp: 'stephaneducasse 2/4/2006 20:41'!
autoScrollForX: x
	"Scroll by the amount x lies outside of my innerBounds.  Return true if this happens."

	| d ticks |
	((d := x - self innerBounds right) > 0
		or: [(d := x - self innerBounds left) < 0])
		ifTrue: [ticks := (self timeForX: self bounds center x + d+1)
						min: score durationInTicks max: 0.
				self moveCursorToTime: ticks.
				scorePlayer ticksSinceStart: ticks.
				^ true].
	^ false
! !

!PianoRollScoreMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
beatLinesOnOff

	showBeatLines := showBeatLines not.
	self changed! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'di 6/20/1999 16:25'!
beatsPerMeasure: n

	^ self timeSignature: n over: notePerBeat! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'stephaneducasse 2/4/2006 20:41'!
contractTime

	timeScale := timeScale / 1.5.
	self rebuildFromScore.
! !

!PianoRollScoreMorph methodsFor: 'editing' stamp: 'dgd 2/21/2003 22:56'!
copySelection
	selection isNil ifTrue: [^self].
	NotePasteBuffer := (score tracks at: selection first) 
				copyFrom: selection second
				to: selection third! !

!PianoRollScoreMorph methodsFor: 'editing' stamp: 'dgd 2/21/2003 22:56'!
cutSelection
	selection isNil ifTrue: [^self].
	self copySelection.
	self deleteSelection! !

!PianoRollScoreMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:38'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!PianoRollScoreMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:29'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color white! !

!PianoRollScoreMorph methodsFor: 'editing' stamp: 'dgd 2/21/2003 22:56'!
deleteSelection
	| selMorphs priorEvent x |
	(selection isNil or: [selection second = 0]) ifTrue: [^self].
	score cutSelection: selection.
	selection second > 1 
		ifTrue: 
			[selection at: 2 put: selection second - 1.
			selection at: 3 put: selection second.
			priorEvent := (score tracks at: selection first) at: selection second.
			(x := self xForTime: priorEvent time) < (self left + 30) 
				ifTrue: [self autoScrollForX: x - ((30 + self width) // 4)]]
		ifFalse: [selection := nil].
	scorePlayer updateDuration.
	self rebuildFromScore.
	selMorphs := self 
				submorphsSatisfying: [:m | (m isKindOf: PianoRollNoteMorph) and: [m selected]].
	selMorphs isEmpty ifFalse: [(selMorphs last noteOfDuration: 0.3) play]! !

!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'stephaneducasse 2/4/2006 20:41'!
drawMeasureLinesOn: aCanvas

	| ticksPerMeas x measureLineColor inner |
	showBeatLines ifNil: [showBeatLines := false].
	showMeasureLines ifNil: [showMeasureLines := true].
	notePerBeat ifNil: [self timeSignature: 4 over: 4].
	showBeatLines ifTrue:
		[measureLineColor := Color gray: 0.8.
		ticksPerMeas := score ticksPerQuarterNote.
		inner := self innerBounds.
		(leftEdgeTime + ticksPerMeas truncateTo: ticksPerMeas)
			to: ((self timeForX: self right - borderWidth) truncateTo: ticksPerMeas)
			by: ticksPerMeas
			do: [:tickTime | x := self xForTime: tickTime.
				aCanvas fillRectangle: (x @ inner top extent: 1 @ inner height)
					color: measureLineColor]].

	showMeasureLines ifTrue:
		[measureLineColor := Color gray: 0.7.
		ticksPerMeas := beatsPerMeasure*score ticksPerQuarterNote*4//notePerBeat.
		inner := self innerBounds.
		(leftEdgeTime + ticksPerMeas truncateTo: ticksPerMeas)
			to: ((self timeForX: self right - borderWidth) truncateTo: ticksPerMeas)
			by: ticksPerMeas
			do: [:tickTime | x := self xForTime: tickTime.
				aCanvas fillRectangle: (x @ inner top extent: 1 @ inner height)
						color: (tickTime = 0 ifTrue: [Color black] ifFalse: [measureLineColor])]].
! !

!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'jm 5/30/1999 17:56'!
drawOn: aCanvas

	super drawOn: aCanvas.
	self drawStaffOn: aCanvas.
! !

!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'stephaneducasse 2/4/2006 20:41'!
drawStaffOn: aCanvas

	| blackKeyColor l r topEdge y |
	self drawMeasureLinesOn: aCanvas.

	blackKeyColor := Color gray: 0.5.
	l := self left + borderWidth.
	r := self right - borderWidth.
	topEdge := self top + borderWidth + 3.
	lowestNote to: 127 do: [:k |
		y := self yForMidiKey: k.
		y <= topEdge ifTrue: [^ self].  "over the top!!"
		(self isBlackKey: k) ifTrue: [
			aCanvas
				fillRectangle: (l@y corner: r@(y + 1))
				color: blackKeyColor]].
! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'stephaneducasse 2/4/2006 20:41'!
expandTime

	timeScale := timeScale * 1.5.
	self rebuildFromScore.
! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'di 6/20/1999 15:25'!
extent: aPoint
	"Force rebuild when re-sized."

	super extent: aPoint. 
	score ifNotNil: [self updateLowestNote].
	self rebuildFromScore.
! !

!PianoRollScoreMorph methodsFor: 'layout' stamp: 'stephaneducasse 2/4/2006 20:41'!
fullBounds
	"Overridden to clip submorph hit detection to my bounds."

	fullBounds ifNil: [fullBounds := bounds].
	^ bounds
! !

!PianoRollScoreMorph methodsFor: 'scrolling' stamp: 'stephaneducasse 2/4/2006 20:41'!
goToTime: scoreTime

	| track trackSize index newLeftEdgeTime |
	newLeftEdgeTime := scoreTime asInteger.
	newLeftEdgeTime < leftEdgeTime
		ifTrue: [indexInTrack := Array new: score tracks size+1 withAll: 1].
	leftEdgeTime := newLeftEdgeTime.
	1 to: score tracks size do: [:trackIndex |
		track := score tracks at: trackIndex.
		index := indexInTrack at: trackIndex.
		trackSize := track size.
		[(index < trackSize) and:
		 [(track at: index) endTime < leftEdgeTime]]
			whileTrue: [index := index + 1].
		indexInTrack at: trackIndex put: index].
	self addNotes.
! !

!PianoRollScoreMorph methodsFor: 'event handling' stamp: 'di 6/17/1999 12:40'!
handlesMouseDown: evt

	^ true! !

!PianoRollScoreMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	self extent: 400 @ 300.
	showMeasureLines := true.
	showBeatLines := false.
	self timeSignature: 4 over: 4.
	self clipSubmorphs: true! !

!PianoRollScoreMorph methodsFor: 'editing' stamp: 'dgd 2/21/2003 22:56'!
insertSelection
	self selection isNil ifTrue: [^self].
	score insertEvents: NotePasteBuffer at: self selection.
	scorePlayer updateDuration.
	self rebuildFromScore! !

!PianoRollScoreMorph methodsFor: 'editing' stamp: 'alain.plantec 2/6/2009 17:19'!
insertTransposed
	| delta transposedNotes list |
	list := (12 to: -12 by: -1) collect: [:i | i printString].
	delta := UIManager default chooseFrom: list values: list title: 'Offset' translated.
 	delta ifNil: [^self].
	transposedNotes := NotePasteBuffer 
				collect: [:note | note copy midiKey: note midiKey + delta].
	selection isNil ifTrue: [^self].
	score insertEvents: transposedNotes at: self selection.
	scorePlayer updateDuration.
	self rebuildFromScore! !

!PianoRollScoreMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
invokeScoreMenu: evt
	"Invoke the score's edit menu."

	| menu subMenu |
	menu := MenuMorph new defaultTarget: self.
	menu addList:
		{{'cut' translated.		#cutSelection}.
		{'copy' translated.		#copySelection}.
		{'paste' translated.		#insertSelection}.
		{'paste...' translated.		#insertTransposed}}.
	menu addLine.
	menu addList:
		{{'legato' translated.		#selectionBeLegato}.
		{'staccato' translated.	#selectionBeStaccato}.
		{'normal' translated.		#selectionBeNormal}}.
	menu addLine.
	menu addList:
		{{'expand time' translated.		#expandTime}.
		{'contract time' translated.		#contractTime}}.
	menu addLine.
	subMenu := MenuMorph new defaultTarget: self.
		(2 to: 12) do: [:i | subMenu add: i printString selector: #beatsPerMeasure: argument: i].
		menu add: 'time   ' translated, beatsPerMeasure printString subMenu: subMenu.
	subMenu := MenuMorph new defaultTarget: self.
		#(2 4 8) do: [:i | subMenu add: i printString selector: #notePerBeat: argument: i].
		menu add: 'sig     ' translated, notePerBeat printString subMenu: subMenu.
	menu addLine.
	showMeasureLines
		ifTrue: [menu add: 'hide measure lines' translated action: #measureLinesOnOff]
		ifFalse: [menu add: 'show measure lines' translated action: #measureLinesOnOff].
	showBeatLines
		ifTrue: [menu add: 'hide beat lines' translated action: #beatLinesOnOff]
		ifFalse: [menu add: 'show beat lines' translated action: #beatLinesOnOff].

	menu addLine.
	menu add: 'add keyboard' translated action: #addKeyboard.

	menu popUpEvent: evt in: self world.
! !

!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'stephaneducasse 2/4/2006 20:41'!
isBlackKey: midiKey
	"Answer true if the given MIDI key corresponds to a black key on the piano keyboard."

	| note |
	note := midiKey \\ 12.
	note = 1 ifTrue: [^ true].
	note = 3 ifTrue: [^ true].
	note = 6 ifTrue: [^ true].
	note = 8 ifTrue: [^ true].
	note = 10 ifTrue: [^ true].
	^ false
! !

!PianoRollScoreMorph methodsFor: 'layout' stamp: 'jm 9/11/1998 09:20'!
layoutChanged
	"Override this to avoid propagating 'layoutChanged' when just adding/removing note objects."

	fullBounds = bounds ifTrue: [^ self].
	super layoutChanged.
! !

!PianoRollScoreMorph methodsFor: 'menu' stamp: 'stephaneducasse 2/4/2006 20:41'!
measureLinesOnOff

	showMeasureLines := showMeasureLines not.
	self changed! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'di 6/15/1999 14:48'!
midiKeyForY: y

	^ lowestNote - ((y - (bounds bottom - borderWidth - 4)) // 3)
! !

!PianoRollScoreMorph methodsFor: 'event handling' stamp: 'stephaneducasse 2/4/2006 20:41'!
mouseDown: evt

	| noteMorphs chordRect sound |
	(self notesInRect: ((evt cursorPoint extent: 1@0) expandBy: 2@30)) isEmpty
		ifTrue: ["If not near a note, then put up score edit menu"
				^ self invokeScoreMenu: evt].

	"Clicked near (but not on) a note, so play all notes at the cursor time"
	noteMorphs := self notesInRect: ((evt cursorPoint extent: 1@0) expandBy: 0@self height).
	chordRect := (self innerBounds withLeft: evt cursorPoint x) withWidth: 1.
	soundsPlayingMorph := Morph newBounds: chordRect color: Color green.
	self addMorphBack: soundsPlayingMorph.
	
	soundsPlaying := IdentityDictionary new.
	noteMorphs do:
		[:m | sound := m soundOfDuration: 999.0.
		soundsPlaying at: m put: sound.
		SoundPlayer resumePlaying: sound quickStart: false].

! !

!PianoRollScoreMorph methodsFor: 'event handling' stamp: 'stephaneducasse 2/4/2006 20:41'!
mouseMove: evt

	| noteMorphs chordRect sound |
	soundsPlaying ifNil: [^ self].
	self autoScrollForX: evt cursorPoint x.

	"Play all notes at the cursor time"
	noteMorphs := self notesInRect: ((evt cursorPoint extent: 1@0) expandBy: 0@self height).
	chordRect := (self innerBounds withLeft: evt cursorPoint x) withWidth: 1.
	soundsPlayingMorph delete.
	soundsPlayingMorph := Morph newBounds: chordRect color: Color green.
	self addMorphBack: soundsPlayingMorph.
	
	noteMorphs do:
		[:m |  "Add any new sounds"
		(soundsPlaying includesKey: m)
			ifFalse: [sound := m soundOfDuration: 999.0.
					soundsPlaying at: m put: sound.
					SoundPlayer resumePlaying: sound quickStart: false]].
	soundsPlaying keys do:
		[:m |  "Remove any sounds no longer in selection."
		(noteMorphs includes: m)
			ifFalse: [(soundsPlaying at: m) stopGracefully.
					soundsPlaying removeKey: m]].

! !

!PianoRollScoreMorph methodsFor: 'event handling' stamp: 'stephaneducasse 2/4/2006 20:41'!
mouseUp: evt

	soundsPlayingMorph ifNotNil: [soundsPlayingMorph delete].
	soundsPlaying ifNotNil: [soundsPlaying do: [:s | s stopGracefully]].
	soundsPlayingMorph := soundsPlaying := nil
! !

!PianoRollScoreMorph methodsFor: 'scrolling' stamp: 'stephaneducasse 2/4/2006 20:41'!
moveCursorToTime: scoreTime

	| cursorOffset desiredCursorHeight |
	scorePlayer isPlaying
		ifTrue:
			[cursorOffset := ((scoreTime - leftEdgeTime) asFloat * timeScale) asInteger.
			(cursorOffset < 0
				or: [cursorOffset > (self width-20)])
				ifTrue:
				[self goToTime: scoreTime - (20/timeScale).
				cursorOffset := ((scoreTime - leftEdgeTime) asFloat * timeScale) asInteger]]
		ifFalse:
			[self goToTime: (scoreTime - (self width//2 / timeScale)
							max: (self width//10 / timeScale) negated).
			cursorOffset := ((scoreTime - leftEdgeTime) asFloat * timeScale) asInteger].

	cursor position: (self left + borderWidth + cursorOffset)@(self top + borderWidth).
	desiredCursorHeight := self height.
	cursor height ~= desiredCursorHeight ifTrue: [cursor extent: 1@desiredCursorHeight].
! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'di 10/19/2000 10:07'!
movieClipPlayer

	^ movieClipPlayer! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:41'!
movieClipPlayer: moviePlayer

	movieClipPlayer := moviePlayer
! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'di 6/20/1999 16:25'!
notePerBeat: n

	^ self timeSignature: beatsPerMeasure over: n! !

!PianoRollScoreMorph methodsFor: 'scrolling' stamp: 'di 6/17/1999 09:36'!
notesInRect: timeSlice

	^ self submorphsSatisfying:
		[:m | (timeSlice intersects: m bounds)
				and: [m isKindOf: PianoRollNoteMorph]]! !

!PianoRollScoreMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
on: aScorePlayer

	scorePlayer := aScorePlayer.
	score := aScorePlayer score.
	colorForTrack := Color wheel: score tracks size.
	leftEdgeTime := 0.
	timeScale := 0.1.
	indexInTrack := Array new: score tracks size withAll: 1.
	lastUpdateTick := -1.

	self updateLowestNote
! !

!PianoRollScoreMorph methodsFor: 'drawing' stamp: 'di 6/20/1999 12:23'!
rebuildFromScore
	"Rebuild my submorphs from the score. This method should be invoked after changing the time scale, the color or visibility of a track, the extent of this morph, etc."

	score ifNil: [^ self].
	self addNotes.
! !

!PianoRollScoreMorph methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:41'!
removedMorph: aMorph
	| trackSize |
	trackSize := score ambientTrack size.
	score removeAmbientEventWithMorph: aMorph.
	trackSize = score ambientTrack size ifFalse:
		["Update duration if we removed an event"
		scorePlayer updateDuration].
	^super removedMorph: aMorph! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'di 6/15/1999 15:46'!
score

	^ score! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'di 6/15/1999 14:52'!
scorePlayer

	^ scorePlayer! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:41'!
selection
	"Returns an array of 3 elements:
		trackIndex
		indexInTrack of first note
		indexInTrack of last note"

	| trackIndex track |
	selection ifNil:  "If no selection, return last event of 1st non-muted track (or nil)"
		[trackIndex := (1 to: score tracks size)
			detect: [:i | (scorePlayer mutedForTrack: i) not] ifNone: [^ nil].
		track := score tracks at: trackIndex.
		^ Array with: trackIndex with: track size with: track size].
	(scorePlayer mutedForTrack: selection first)
		ifTrue: [selection := nil.  ^ self selection].
	^ selection! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:41'!
selection: anArray

	selection := anArray! !

!PianoRollScoreMorph methodsFor: 'stepping and presenter' stamp: 'stephaneducasse 2/4/2006 20:41'!
step

	| t |
	score ifNil: [^ self].

	lastMutedState ~= scorePlayer mutedState ifTrue: [
		self rebuildFromScore.
		lastMutedState := scorePlayer mutedState copy].

	t := scorePlayer ticksSinceStart.
	t = lastUpdateTick ifFalse: [
		self moveCursorToTime: t.
		lastUpdateTick := t].
! !

!PianoRollScoreMorph methodsFor: 'testing' stamp: 'jm 6/1/1998 09:07'!
stepTime

	^ 0
! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'di 10/12/2000 00:40'!
tickTimeAtCursor
	cursor ifNil: [^ 0].
	^ self timeForX: cursor left! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'di 8/3/1998 21:35'!
timeForX: aNumber

	^ ((aNumber - bounds left - borderWidth) asFloat / timeScale + leftEdgeTime) asInteger! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'di 10/19/2000 21:19'!
timeScale

	^ timeScale  "in pixels per tick"! !

!PianoRollScoreMorph methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:41'!
timeSignature: num over: denom

	beatsPerMeasure := num.
	notePerBeat := denom.  "a number like 2, 4, 8"
	self changed! !

!PianoRollScoreMorph methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:41'!
updateLowestNote
	"find the actual lowest note in the score"

	| n |
	lowestNote := 128 - (self innerBounds height // 3).
	score tracks do: [:track |
		1 to: track size do: [:i |
			n := track at: i.
			(n isNoteEvent and: [n midiKey < lowestNote])
				ifTrue: [lowestNote := n midiKey - 4]]].
! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'di 8/3/1998 21:29'!
xForTime: aNumber

	^ ((aNumber - leftEdgeTime) asFloat * timeScale) asInteger + bounds left + borderWidth
! !

!PianoRollScoreMorph methodsFor: 'geometry' stamp: 'jm 7/10/1998 14:35'!
yForMidiKey: midiKey

	^ (bounds bottom - borderWidth - 4) - (3 * (midiKey - lowestNote))
! !

RectangleMorph subclass: #SoundLoopMorph
	instanceVariableNames: 'samplesUntilNextControl seqSound cursor controlIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SoundLoopMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:21'!
addCursorMorph
	self addMorph:
		(cursor := (RectangleMorph
				newBounds: (self innerBounds topLeft extent: 1@self innerBounds height)
				color: Color red)
						borderWidth: 0)! !

!SoundLoopMorph methodsFor: 'as yet unclassified' stamp: 'jm 1/5/98 17:31'!
buildSound
	"Build a compound sound for the next iteration of the loop."

	| mixer soundMorphs startTime pan |
	mixer := MixedSound new.
	mixer add: (RestSound dur: (self width - (2 * borderWidth)) / 128.0).
	soundMorphs := self submorphs select: [:m | m respondsTo: #sound].
	soundMorphs do: [:m |
		startTime := (m position x - (self left + borderWidth)) / 128.0.
		pan := (m position y - (self top + borderWidth)) asFloat / (self height - (2 * borderWidth) - m height).
		mixer add: ((RestSound dur: startTime), m sound copy) pan: pan].
	^ mixer
! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 11/14/97 11:21'!
controlRate
	"Answer the number of control changes per second."

	^ 32
! !

!SoundLoopMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:40'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!SoundLoopMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:07'!
defaultBounds
"answer the default bounds for the receiver"
	^ 0 @ 0 corner: 128 @ 128 + (self defaultBorderWidth * 2)! !

!SoundLoopMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color lightBlue! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 11/14/97 11:21'!
doControl

	seqSound doControl.
	controlIndex := controlIndex + 1.
	controlIndex >= (self controlRate * (self innerBounds width // 128))
		ifTrue: [controlIndex := 0].
! !

!SoundLoopMorph methodsFor: 'geometry' stamp: 'jm 11/14/97 11:21'!
extent: newExtent
	super extent: (newExtent truncateTo: 128@128) + (self borderWidth*2)! !

!SoundLoopMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:50'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	controlIndex := 0.
	self addCursorMorph! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 1/5/98 13:40'!
mixSampleCount: n into: aSoundBuffer startingAt: startIndex leftVol: leftVol rightVol: rightVol
	"Repeatedly play my sounds."

	| i count samplesNeeded |
	i := startIndex.
	samplesNeeded := n.
	[samplesNeeded > 0] whileTrue: [
		count := seqSound samplesRemaining min: samplesNeeded.
		count = 0 ifTrue: [
			self reset.
			count := seqSound samplesRemaining min: samplesNeeded.
			count = 0 ifTrue: [^ self]].  "zero length sound"
		seqSound mixSampleCount: count into: aSoundBuffer startingAt: i leftVol: leftVol rightVol: rightVol.
		i := i + count.
		samplesNeeded := samplesNeeded - count].
! !

!SoundLoopMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:11'!
play
	"Play this sound to the sound ouput port in real time."

	self reset.
	SoundPlayer playSound: self.
! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 1/26/98 22:05'!
playSampleCount: n into: aSoundBuffer startingAt: startIndex
	"Mixes the next count samples of this sound into the given buffer starting at the given index, updating the receiver's control parameters at periodic intervals."

	| fullVol samplesBetweenControlUpdates pastEnd i remainingSamples count |
	fullVol := AbstractSound scaleFactor.
	samplesBetweenControlUpdates := self samplingRate // self controlRate.
	pastEnd := startIndex + n.  "index just index of after last sample"
	i := startIndex.
	[i < pastEnd] whileTrue: [
		remainingSamples := self samplesRemaining.
		remainingSamples <= 0 ifTrue: [^ self].
		count := pastEnd - i.
		samplesUntilNextControl < count ifTrue: [count := samplesUntilNextControl].
		remainingSamples < count ifTrue: [count := remainingSamples].
		self mixSampleCount: count into: aSoundBuffer startingAt: i leftVol: fullVol rightVol: fullVol.
		samplesUntilNextControl := samplesUntilNextControl - count.
		samplesUntilNextControl <= 0 ifTrue: [
			self doControl.
			samplesUntilNextControl := samplesBetweenControlUpdates].
		i := i + count].
! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 11/14/97 11:21'!
positionCursor
	| x |
	x := controlIndex * 128 // self controlRate.
	cursor position: self innerBounds topLeft + (x@0)
! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 11/14/97 11:21'!
reset
	"Reset my internal state for a replay."

	seqSound := self buildSound reset.
	samplesUntilNextControl := (self samplingRate // self controlRate).
	controlIndex := 0.
	self positionCursor! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 11/14/97 11:11'!
samplesRemaining

	^ 1000000
! !

!SoundLoopMorph methodsFor: 'playing' stamp: 'jm 11/14/97 11:11'!
samplingRate
	"Answer the sampling rate in samples per second."

	^ SoundPlayer samplingRate! !

!SoundLoopMorph methodsFor: 'stepping and presenter' stamp: 'jm 11/14/97 11:21'!
step
	self positionCursor! !

!SoundLoopMorph methodsFor: 'testing' stamp: 'jm 11/14/97 11:21'!
stepTime

	^ 50
! !

!SoundLoopMorph methodsFor: 'stepping and presenter' stamp: 'jm 11/14/97 11:11'!
stop
	"Stop playing this sound."

	SoundPlayer pauseSound: self.
! !

!SoundLoopMorph methodsFor: 'dropping/grabbing' stamp: 'jm 11/14/97 11:11'!
wantsDroppedMorph: aMorph event: evt

	^ aMorph respondsTo: #sound
! !

RectangleMorph subclass: #SoundSequencerMorph
	instanceVariableNames: 'controlPanel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SoundSequencerMorph methodsFor: 'initialization' stamp: 'jm 11/14/97 11:21'!
initialize
	super initialize.
	self extent: 550@350.
	self makeControlPanel.
	self addMorph: controlPanel.
	self addMorph: ((SoundLoopMorph newBounds: (10@40 extent: 128@128)) extent: 128@128).
	self addMorph: ((SoundLoopMorph newBounds: (10@200 extent: 512@128)) extent: 512@128).! !

!SoundSequencerMorph methodsFor: 'as yet unclassified' stamp: 'tk 2/19/2001 18:14'!
makeControlPanel
	| bb cc |
	cc := Color black.
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	controlPanel := AlignmentMorph newRow.
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	controlPanel color: bb color; borderWidth: 0; layoutInset: 0.
	controlPanel hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	controlPanel addMorphBack: (bb label: 'reset';	actionSelector: #reset).
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	controlPanel addMorphBack: (bb label: 'stop';		actionSelector: #stop).
	bb := SimpleButtonMorph new target: self; borderColor: cc.
	controlPanel addMorphBack: (bb label: 'play';	actionSelector: #play).
! !

!SoundSequencerMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:08'!
play
	self submorphsDo: [:m | m == controlPanel ifFalse: [m play]]! !

!SoundSequencerMorph methodsFor: 'as yet unclassified' stamp: 'jm 11/14/97 11:08'!
reset
	self submorphsDo: [:m | m == controlPanel ifFalse: [m reset]]! !

!SoundSequencerMorph methodsFor: 'stepping and presenter' stamp: 'jm 11/14/97 11:21'!
stop
	self submorphsDo: [:m | m == controlPanel ifFalse: [m stop]].
	SoundPlayer shutDown! !

Morph subclass: #SpeakerMorph
	instanceVariableNames: 'bufferSize buffer lastConePosition sound'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SoundMorphicUserInterface'!

!SpeakerMorph methodsFor: 'initialization' stamp: 'nk 6/12/2004 10:05'!
addGraphic

	| graphic |
	graphic := World drawingClass withForm: self speakerGraphic.
	graphic position: bounds center - (graphic extent // 2).
	self addMorph: graphic.
! !

!SpeakerMorph methodsFor: 'nil' stamp: 'jm 4/22/1999 16:45'!
appendSample: aFloat
	"Append the given sample, a number between -100.0 and 100.0, to my buffer. Flush the buffer if it is full."

	lastConePosition := aFloat.
	lastConePosition > 100.0 ifTrue: [lastConePosition := 100.0].
	lastConePosition < -100.0 ifTrue: [lastConePosition := -100.0].
	buffer nextPut: (327.67 * lastConePosition) truncated.
	buffer position >= bufferSize ifTrue: [self flushBuffer].
! !

!SpeakerMorph methodsFor: 'speaker' stamp: 'jm 4/21/1999 10:18'!
conePosition

	^ lastConePosition
! !

!SpeakerMorph methodsFor: 'speaker' stamp: 'jm 4/22/1999 16:46'!
conePosition: aNumber

	self appendSample: aNumber asFloat.  "sets lastConePosition"
! !

!SpeakerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:51'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color
		r: 1.0
		g: 0.484
		b: 0.258! !

!SpeakerMorph methodsFor: 'speaker' stamp: 'jm 4/22/1999 13:29'!
flushBuffer

	| buf |
	buf := buffer contents.
	buffer resetContents.
	sound isPlaying ifFalse: [sound := SequentialSound new].
	sound add: (SampledSound samples: buf samplingRate: 11025).
	sound isPlaying
		ifTrue: [sound pruneFinishedSounds]
		ifFalse: [sound play].
! !

!SpeakerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:51'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	self addGraphic.
	bufferSize := 5000.
	buffer := WriteStream
				on: (SoundBuffer newMonoSampleCount: bufferSize).
	lastConePosition := 0.
	sound := SequentialSound new! !

!SpeakerMorph methodsFor: 'initialization' stamp: 'jm 4/22/1999 22:58'!
speakerGraphic

	^ Form
		extent: 19@18
		depth: 8
		fromArray: #(0 0 1493172224 2816 0 0 0 1493172224 11 0 0 138 1493172224 184549376 184549376 0 35509 2315255808 720896 720896 0 9090522 2315255808 2816 720896 0 2327173887 2315255819 2816 720896 138 3051028442 2315255819 2816 2816 1505080590 4294957786 2315255808 184549387 2816 3053453311 4292532917 1493172224 184549387 2816 1505080714 3048584629 1493172224 184549387 2816 9079434 3048584629 1493172224 184549387 2816 138 2327164341 1493172235 2816 2816 0 2324346293 1493172235 2816 720896 0 9079477 1493172224 2816 720896 0 35466 1493172224 720896 720896 0 138 0 184549376 184549376 0 0 0 11 0 0 0 0 2816 0)
		offset: 0@0
! !

!SpeakerMorph methodsFor: 'speaker' stamp: 'jm 4/22/1999 15:33'!
stopSound

	sound pause.
	buffer resetContents.
! !
ScorePlayerMorph initialize!
